import importlib
import json
import os
import sys
import tempfile

from confluent_kafka.schema_registry.avro import AvroSerializer
from confluent_kafka.schema_registry.json_schema import JSONSerializer
from confluent_kafka.schema_registry.protobuf import ProtobufSerializer
from confluent_kafka.serialization import MessageField, SerializationContext

from google.protobuf.json_format import ParseDict

from kafi.schemaregistry import SchemaRegistry
from kafi.helpers import to_bytes

class Serializer(SchemaRegistry):
    def __init__(self, schema_registry_config_dict):
        super().__init__(schema_registry_config_dict)

    def serialize(self, payload, key_bool, normalize_schemas=False):
        type_str = self.key_type_str if key_bool else self.value_type_str
        schema_str_or_dict = self.key_schema_str_or_dict if key_bool else self.value_schema_str_or_dict
        schema_id_int = self.key_schema_id_int if key_bool else self.value_schema_id_int
        messageField = MessageField.KEY if key_bool else MessageField.VALUE
        #
        def get_schema_str():
            if schema_str_or_dict is None:
                if schema_id_int is None:
                    raise Exception("Please provide a schema or schema ID for the " + ("key" if key_bool else "value") + ".")
                schema_str = self.schemaRegistryClient.get_schema(schema_id_int)
            else:
                if isinstance(schema_str_or_dict, str):
                    schema_str = schema_str_or_dict
                elif isinstance(schema_str_or_dict, dict):
                    schema_str = json.dumps(schema_str_or_dict)
            #
            return schema_str
        #

        def payload_to_payload_dict():
            if isinstance(payload, bytes):
                payload_dict = json.loads(payload)
            elif isinstance(payload, str):
                payload_dict = json.loads(payload)
            elif isinstance(payload, dict):
                payload_dict = payload
            #
            return payload_dict
        #
        if payload == None:
            serialized_payload_bytes = None
        else:
            if type_str.lower() in ["bytes", "str", "json"]:
                serialized_payload_bytes = to_bytes(payload)
            elif type_str.lower() in ["pb", "protobuf"]:
                schema = get_schema_str()
                generalizedProtocolMessageType = self.schema_str_to_generalizedProtocolMessageType(schema, self.topic_str, key_bool, normalize_schemas)
                protobufSerializer = ProtobufSerializer(generalizedProtocolMessageType, self.schemaRegistryClient, {"use.deprecated.format": False})
                payload_dict = payload_to_payload_dict()
                protobuf_message = generalizedProtocolMessageType()
                ParseDict(payload_dict, protobuf_message)
                serialized_payload_bytes = protobufSerializer(protobuf_message, SerializationContext(self.topic_str, messageField))
            elif type_str.lower() == "avro":
                schema = get_schema_str()
                avroSerializer = AvroSerializer(self.schemaRegistryClient, schema)
                payload_dict = payload_to_payload_dict()
                serialized_payload_bytes = avroSerializer(payload_dict, SerializationContext(self.topic_str, messageField))
            elif type_str.lower() in ["jsonschema", "json_sr"]:
                payload_dict = payload_to_payload_dict()
                schema = get_schema_str()
                jSONSerializer = JSONSerializer(schema, self.schemaRegistryClient)
                serialized_payload_bytes = jSONSerializer(payload_dict, SerializationContext(self.topic_str, messageField))
            else:
                raise Exception("Only \"bytes\", \"str\", \"json\", \"avro\", \"protobuf\" (\"pb\") and \"jsonschema\" (\"json_sr\") supported.")
        #
        return serialized_payload_bytes

    # Helpers

    def schema_str_to_generalizedProtocolMessageType(self, schema_str, topic_str, key_bool, normalize_schemas=False):
        schema_hash_int = hash(schema_str)
        if schema_hash_int in self.schema_hash_int_generalizedProtocolMessageType_dict:
            generalizedProtocolMessageType = self.schema_hash_int_generalizedProtocolMessageType_dict[schema_hash_int]
        else:
            subject_name_str = self.create_subject_name_str(topic_str, key_bool)
            schema_dict = self.create_schema_dict(schema_str, "PROTOBUF")
            schema_id_int = self.register_schema(subject_name_str, schema_dict, normalize_schemas)
            #
            generalizedProtocolMessageType = self.schema_id_int_and_schema_str_to_generalizedProtocolMessageType(schema_id_int, schema_str)
            #
            self.schema_hash_int_generalizedProtocolMessageType_dict[schema_hash_int] = generalizedProtocolMessageType
        #
        return generalizedProtocolMessageType

    def schema_id_int_and_schema_str_to_generalizedProtocolMessageType(self, schema_id_int, schema_str):
        path_str = f"/{tempfile.gettempdir()}/kafi/clusters/{self.storage_obj.config_str}"
        os.makedirs(path_str, exist_ok=True)
        file_str = f"schema_{schema_id_int}.proto"
        file_path_str = f"{path_str}/{file_str}"
        with open(file_path_str, "w") as textIOWrapper:
            textIOWrapper.write(schema_str)
        #
        import grpc_tools.protoc
        grpc_tools.protoc.main(["protoc", f"-I{path_str}", f"--python_out={path_str}", f"{file_str}"])
        #
        sys.path.insert(1, path_str)
        schema_module = importlib.import_module(f"schema_{schema_id_int}_pb2")
        schema_name_str = list(schema_module.DESCRIPTOR.message_types_by_name.keys())[0]
        generalizedProtocolMessageType = getattr(schema_module, schema_name_str)
        return generalizedProtocolMessageType



import time
from kafi.helpers import zip2

# Constants

ALL_MESSAGES = -1

#

class Functional:
    def foldl(self, topic, foldl_function, initial_acc, n=ALL_MESSAGES, **kwargs):
        verbose_int = self.verbose()
        #
        progress_num_messages_int = self.progress_num_messages()
        #
        consumer = self.consumer(topic, **kwargs)
        #
        def foldl_function1(acc_consume_message_counter_int_tuple, message_dict):
            (acc, consume_message_counter_int) = acc_consume_message_counter_int_tuple
            #
            acc = foldl_function(acc, message_dict)
            #
            consume_message_counter_int += 1
            if verbose_int > 0 and consume_message_counter_int % progress_num_messages_int == 0:
                print(f"Read: {consume_message_counter_int}")
            #
            return (acc, consume_message_counter_int)
        #
        acc_consume_message_counter_int_tuple = consumer.foldl(foldl_function1, (initial_acc, 0), n, **kwargs)
        #
        consumer.close()
        #
        return acc_consume_message_counter_int_tuple

    #

    def flatmap(self, topic, flatmap_function, n=ALL_MESSAGES, **kwargs):
        def foldl_function(list, message_dict):
            list += flatmap_function(message_dict)
            #
            return list
        #
        return self.foldl(topic, foldl_function, [], n, **kwargs)

    def map(self, topic, map_function, n=ALL_MESSAGES, **kwargs):
        def flatmap_function(message_dict):
            return [map_function(message_dict)]
        #
        return self.flatmap(topic, flatmap_function, n, **kwargs)

    def filter(self, topic, filter_function, n=ALL_MESSAGES, **kwargs):
        def flatmap_function(message_dict):
            return [message_dict] if filter_function(message_dict) else []
        #
        return self.flatmap(topic, flatmap_function, n, **kwargs)

    def foreach(self, topic, foreach_function, n=ALL_MESSAGES, **kwargs):
        def foldl_function(_, message_dict):
            foreach_function(message_dict)
        #
        self.foldl(topic, foldl_function, None, n, **kwargs)

    #

    def foldl_to(self, topic, target_storage, target_topic, foldl_to_function, initial_acc, n=ALL_MESSAGES, **kwargs):
        verbose_int = self.verbose()
        #
        progress_num_messages_int = self.progress_num_messages()
        #

        def foldl_to_function1(acc_consume_message_counter_int_produce_batch_size_int_produce_batch_message_dict_list_produce_message_counter_int_tuple, message_dict):
            (acc, consume_message_counter_int, produce_batch_size_int, produce_batch_message_dict_list, produce_message_counter_int) = acc_consume_message_counter_int_produce_batch_size_int_produce_batch_message_dict_list_produce_message_counter_int_tuple
            #
            (acc, message_dict_list) = foldl_to_function(acc, message_dict)
            #
            consume_message_counter_int += 1
            if verbose_int > 0 and consume_message_counter_int % progress_num_messages_int == 0:
                print(f"Read: {consume_message_counter_int}")
            #
            produce_batch_message_dict_list += message_dict_list
            #
            if len(produce_batch_message_dict_list) == produce_batch_size_int:
                target_producer.produce_list(produce_batch_message_dict_list, **target_kwargs)
                #
                produce_message_counter_int += len(produce_batch_message_dict_list)
                if verbose_int > 0 and produce_message_counter_int % progress_num_messages_int == 0:
                    print(f"Written: {produce_message_counter_int}")
                #
                return (acc, consume_message_counter_int, produce_batch_size_int, [], produce_message_counter_int)
            else:
                return (acc, consume_message_counter_int, produce_batch_size_int, produce_batch_message_dict_list, produce_message_counter_int)

        #
        source_kwargs = self.copy_kwargs("source", **kwargs)
        #
        target_kwargs = self.copy_kwargs("target", **kwargs)
        #
        produce_batch_size_int = kwargs["produce_batch_size"] if "produce_batch_size" in kwargs else target_storage.produce_batch_size()
        #
        consumer = self.consumer(topic, **source_kwargs)
        #
        target_producer = target_storage.producer(target_topic, **target_kwargs)
        #
        (acc, consume_message_counter_int, _, produce_batch_message_dict_list, produce_message_counter_int) = consumer.foldl(foldl_to_function1, (initial_acc, 0, produce_batch_size_int, [], 0), n, **kwargs)
        #
        consumer.close()
        #
        if len(produce_batch_message_dict_list) > 0:
            target_producer.produce_list(produce_batch_message_dict_list, **target_kwargs)
            produce_message_counter_int += len(produce_batch_message_dict_list)
        #
        target_producer.close()
        #
        return (acc, consume_message_counter_int, produce_message_counter_int)

    def flatmap_to(self, topic, target_storage, target_topic, flatmap_function, n=ALL_MESSAGES, **kwargs):
        def foldl_to_function(_, message_dict):
            return (None, flatmap_function(message_dict))
        #
        (_, consume_message_counter_int, produce_message_counter_int) = self.foldl_to(topic, target_storage, target_topic, foldl_to_function, None, n, **kwargs)
        return (consume_message_counter_int, produce_message_counter_int)

    def map_to(self, topic, target_storage, target_topic, map_function, n=ALL_MESSAGES, **kwargs):
        def flatmap_function(message_dict):
            return [map_function(message_dict)]
        #
        return self.flatmap_to(topic, target_storage, target_topic, flatmap_function, n, **kwargs)

    def filter_to(self, topic, target_storage, target_topic, filter_function, n=ALL_MESSAGES, **kwargs):
        def flatmap_function(message_dict):
            return [message_dict] if filter_function(message_dict) else []
        #
        return self.flatmap_to(topic, target_storage, target_topic, flatmap_function, n, **kwargs)

    #

    def zip_foldl(self, source1_topic, source2_storage, source2_topic, zip_foldl_function, initial_acc, n=ALL_MESSAGES, **kwargs):
        n_int = n
        #
        verbose_int = self.verbose()
        progress_num_messages_int = self.progress_num_messages()
        #
        consume_batch_size_int = kwargs["consume_batch_size"] if "consume_batch_size" in kwargs else self.consume_batch_size()
        if consume_batch_size_int > n_int:
            consume_batch_size_int = n_int
        #
        break_function = kwargs["break_function"] if "break_function" in kwargs else lambda _, _1: False
        #
        source1_kwargs = self.copy_kwargs("source1", **kwargs)
        source2_kwargs = self.copy_kwargs("source2", **kwargs)
        #
        consumer1 = self.consumer(source1_topic, **source1_kwargs)
        consumer2 = source2_storage.consumer(source2_topic, **source2_kwargs)
        #
        message_counter_int1 = 0
        message_counter_int2 = 0
        acc = initial_acc
        break_bool = False
        while True:
            message_dict_list1 = []
            while True:
                message_dict_list1 += consumer1.consume(n=consume_batch_size_int)
                if not message_dict_list1 or consume_batch_size_int == ALL_MESSAGES or len(message_dict_list1) == consume_batch_size_int:
                    break
            if not message_dict_list1:
                break
            num_messages_int1 = len(message_dict_list1)
            message_counter_int1 += num_messages_int1
            if verbose_int > 0 and message_counter_int1 % progress_num_messages_int == 0:
                print(f"Read (storage 1): {message_counter_int1}")
            #
            consume_batch_size_int2 = num_messages_int1 if num_messages_int1 < consume_batch_size_int else consume_batch_size_int
            message_dict_list2 = []
            while True:
                message_dict_list2 += consumer2.consume(n=consume_batch_size_int2)
                if not message_dict_list2 or consume_batch_size_int2 == ALL_MESSAGES or len(message_dict_list2) == consume_batch_size_int2:
                    break
            if not message_dict_list2:
                break
            num_messages_int2 = len(message_dict_list2)
            message_counter_int2 += num_messages_int2
            if verbose_int > 0 and message_counter_int2 % progress_num_messages_int == 0:
                print(f"Read (storage 2): {message_counter_int2}")
            #
            for message_dict1, message_dict2 in zip2(message_dict_list1, message_dict_list2):
                if break_function(message_dict1, message_dict2):
                    break_bool = True
                    break
                acc = zip_foldl_function(acc, message_dict1, message_dict2)
            #
            if break_bool:
                break
            #
            if n_int != ALL_MESSAGES:
                if message_counter_int1 >= n_int:
                    break
        #
        consumer1.close()
        consumer2.close()
        #
        return acc, message_counter_int1, message_counter_int2

    #

    def zip_foldl_to(self, source1_topic, source2_storage, source2_topic, target_storage, target_topic, zip_foldl_to_function, initial_acc, n=ALL_MESSAGES, **kwargs):
        n_int = n
        #
        verbose_int = self.verbose()
        progress_num_messages_int = self.progress_num_messages()
        #
        consume_batch_size_int = kwargs["consume_batch_size"] if "consume_batch_size" in kwargs else self.consume_batch_size()
        if consume_batch_size_int > n_int:
            consume_batch_size_int = n_int
        #
        break_function = kwargs["break_function"] if "break_function" in kwargs else lambda _, _1: False
        #
        source1_kwargs = self.copy_kwargs("source1", **kwargs)
        source2_kwargs = self.copy_kwargs("source2", **kwargs)
        #
        target_kwargs = self.copy_kwargs("target", **kwargs)
        #
        produce_batch_size_int = kwargs["produce_batch_size"] if "produce_batch_size" in kwargs else target_storage.produce_batch_size()
        #
        consumer1 = self.consumer(source1_topic, **source1_kwargs)
        consumer2 = source2_storage.consumer(source2_topic, **source2_kwargs)
        #
        target_producer = target_storage.producer(target_topic, **target_kwargs)
        #
        message_counter_int1 = 0
        message_counter_int2 = 0
        acc = initial_acc
        break_bool = False
        produce_batch_message_dict_list = []
        produce_message_counter_int = 0
        while True:
            message_dict_list1 = []
            while True:
                message_dict_list1 += consumer1.consume(n=consume_batch_size_int)
                if not message_dict_list1 or consume_batch_size_int == ALL_MESSAGES or len(message_dict_list1) == consume_batch_size_int:
                    break
            if not message_dict_list1:
                break
            num_messages_int1 = len(message_dict_list1)
            message_counter_int1 += num_messages_int1
            if verbose_int > 0 and message_counter_int1 % progress_num_messages_int == 0:
                print(f"Read (storage 1): {message_counter_int1}")
            #
            consume_batch_size_int2 = num_messages_int1 if num_messages_int1 < consume_batch_size_int else consume_batch_size_int
            message_dict_list2 = []
            while True:
                message_dict_list2 += consumer2.consume(n=consume_batch_size_int2)
                if not message_dict_list2 or consume_batch_size_int2 == ALL_MESSAGES or len(message_dict_list2) == consume_batch_size_int2:
                    break
            if not message_dict_list2:
                break
            num_messages_int2 = len(message_dict_list2)
            message_counter_int2 += num_messages_int2
            if verbose_int > 0 and message_counter_int2 % progress_num_messages_int == 0:
                print(f"Read (storage 2): {message_counter_int2}")
            #
            collected_message_dict_list = []
            for message_dict1, message_dict2 in zip2(message_dict_list1, message_dict_list2):
                if break_function(message_dict1, message_dict2):
                    break_bool = True
                    break
                (acc, message_dict_list) = zip_foldl_to_function(acc, message_dict1, message_dict2)
                collected_message_dict_list += message_dict_list
            #
            if break_bool:
                break
            #
            produce_batch_message_dict_list += collected_message_dict_list
            #
            if len(produce_batch_message_dict_list) == produce_batch_size_int:
                target_producer.produce_list(produce_batch_message_dict_list, **target_kwargs)
                #
                produce_message_counter_int += len(produce_batch_message_dict_list)
                if verbose_int > 0 and produce_message_counter_int % progress_num_messages_int == 0:
                    print(f"Written: {produce_message_counter_int}")
                #
                produce_batch_message_dict_list = []
            #
            if n_int != ALL_MESSAGES:
                if message_counter_int1 >= n_int:
                    break
        #
        consumer1.close()
        consumer2.close()
        #
        if len(produce_batch_message_dict_list) > 0:
            target_producer.produce_list(produce_batch_message_dict_list, **target_kwargs)
            #
            produce_message_counter_int += len(produce_batch_message_dict_list)
        #
        target_producer.close()
        #
        return acc, message_counter_int1, message_counter_int2, produce_message_counter_int

    #

    def copy_kwargs(self, name_str, **kwargs):
        copied_kwargs = kwargs.copy()
        #
        if f"{name_str}_group" in kwargs:
            copied_kwargs["group"] = kwargs[f"{name_str}_group"]
        if f"{name_str}_offsets" in kwargs:
            copied_kwargs["offsets"] = kwargs[f"{name_str}_offsets"]
        if f"{name_str}_key_type" in kwargs:
            copied_kwargs["key_type"] = kwargs[f"{name_str}_key_type"]
        if f"{name_str}_value_type" in kwargs:
            copied_kwargs["value_type"] = kwargs[f"{name_str}_value_type"]
        if f"{name_str}_type" in kwargs:
            copied_kwargs["type"] = kwargs[f"{name_str}_type"]
        if f"{name_str}_key_schema" in kwargs:
            copied_kwargs["key_schema"] = kwargs[f"{name_str}_key_schema"]
        if f"{name_str}_value_schema" in kwargs:
            copied_kwargs["value_schema"] = kwargs[f"{name_str}_value_schema"]
        if f"{name_str}_key_schema_id" in kwargs:
            copied_kwargs["key_schema_id"] = kwargs[f"{name_str}_key_schema_id"]
        if f"{name_str}_value_schema_id" in kwargs:
            copied_kwargs["value_schema_id"] = kwargs[f"{name_str}_value_schema_id"]
        #
        return copied_kwargs



from kafi.fs.fs_consumer import FSConsumer

#

class S3Consumer(FSConsumer):
    def __init__(self, s3_obj, topic, **kwargs):
        super().__init__(s3_obj, topic, **kwargs)

    #

    def close(self):
        return self.topic_str_list



import os

from kafi.storage_producer import StorageProducer
from kafi.helpers import get_millis

# Constants

CURRENT_TIME = 0
RD_KAFKA_PARTITION_UA = -1
TIMESTAMP_CREATE_TIME = 1

#

class FSProducer(StorageProducer):
    def __init__(self, fs_obj, topic, **kwargs):
        super().__init__(fs_obj, topic, **kwargs)
        #
        if not fs_obj.exists(self.topic_str):
            fs_obj.create(self.topic_str)

    #

    def produce(self, value, **kwargs):
        partition_int_offsets_tuple_dict = self.storage_obj.watermarks(self.topic_str)[self.topic_str]
        last_offsets_dict = {partition_int: offsets_tuple[1] for partition_int, offsets_tuple in partition_int_offsets_tuple_dict.items()}
        #
        key = kwargs["key"] if "key" in kwargs else None
        partition = kwargs["partition"] if "partition" in kwargs else RD_KAFKA_PARTITION_UA
        timestamp = kwargs["timestamp"] if "timestamp" in kwargs else CURRENT_TIME
        headers = kwargs["headers"] if "headers" in kwargs else None
        #
        value_list = value if isinstance(value, list) else [value]
        #
        key_list = key if isinstance(key, list) else [key for _ in value_list]
        #
        partition_int_list = partition if isinstance(partition, list) else [partition for _ in value_list]
        #
        timestamp_list = timestamp if isinstance(timestamp, list) else [timestamp for _ in value_list]
        #
        headers_list = headers if isinstance(headers, list) and all(self.storage_obj.is_headers(headers1) for headers1 in headers) and len(headers) == len(value_list) else [headers for _ in value_list]
        headers_str_bytes_tuple_list_list = [self.storage_obj.headers_to_headers_str_bytes_tuple_list(headers) for headers in headers_list]
        #
        partitions_int = self.storage_obj.admin.get_partitions(self.topic_str)
        partition_int_message_dict_list_dict = {partition_int: [] for partition_int in range(partitions_int)}
        round_robin_counter_int = 0
        partition_int_offset_counter_int_dict = {partition_int: last_offset_int if last_offset_int > 0 else 0 for partition_int, last_offset_int in last_offsets_dict.items()}
        #
        for value, key, timestamp, headers_str_bytes_tuple_list, partition_int in zip(value_list, key_list, timestamp_list, headers_str_bytes_tuple_list_list, partition_int_list):
            if partition_int is RD_KAFKA_PARTITION_UA:
                if key is None:
                    partition_int = round_robin_counter_int
                    if round_robin_counter_int == partitions_int - 1:
                        round_robin_counter_int = 0
                    else:
                        round_robin_counter_int += 1
                else:
                    partition_int = hash(str(key)) % partitions_int
            #
            if timestamp == CURRENT_TIME:
                timestamp = (TIMESTAMP_CREATE_TIME, get_millis())
            #
            message_dict = {"topic": self.topic_str, "value": value, "key": key, "timestamp": timestamp, "headers": headers_str_bytes_tuple_list, "partition": partition_int, "offset": partition_int_offset_counter_int_dict[partition_int]}
            #
            partition_int_message_dict_list_dict[partition_int].append(message_dict)
            #
            partition_int_offset_counter_int_dict[partition_int] += 1
        #
        topic_abs_dir_str = self.storage_obj.admin.get_topic_abs_path_str(self.topic_str)
        for partition_int, message_dict_list in partition_int_message_dict_list_dict.items():
            if len(message_dict_list) > 0:
                start_timestamp_int = message_dict_list[0]["timestamp"][1]
                end_timestamp_int = message_dict_list[-1]["timestamp"][1]
                #
                messages_bytes = b""
                for message_dict in message_dict_list:
                    message_dict["key"] = self.serialize(message_dict["key"], True)
                    message_dict["value"] = self.serialize(message_dict["value"], False)
                    #
                    start_offset_int = last_offsets_dict[partition_int]
                    end_offset_int = start_offset_int + len(message_dict_list) - 1
                    #
                    abs_path_file_str = os.path.join(topic_abs_dir_str, "partitions", f"{partition_int:09},{start_offset_int:021},{end_offset_int:021},{start_timestamp_int},{end_timestamp_int}")
                    #
                    message_bytes = str(message_dict).encode("utf-8") + b"\n"
                    #
                    messages_bytes += message_bytes
                #
                self.storage_obj.admin.write_bytes(abs_path_file_str, messages_bytes)



from confluent_kafka import Consumer, TopicPartition

from kafi.kafka.kafka_consumer import KafkaConsumer

# Constants

ALL_MESSAGES = -1

#

class ClusterConsumer(KafkaConsumer):
    def __init__(self, cluster_obj, *topics, **kwargs):
        super().__init__(cluster_obj, *topics, **kwargs)
        #
        # Consumer config
        #
        self.consumer_config_dict.update(cluster_obj.kafka_config_dict)
        #
        if "group.id" not in self.consumer_config_dict:
            self.consumer_config_dict["group.id"] = self.group_str
        if "session.timeout.ms" not in self.consumer_config_dict:
            self.consumer_config_dict["session.timeout.ms"] = cluster_obj.session_timeout_ms()
        if "enable.auto.commit" not in self.consumer_config_dict:
            self.consumer_config_dict["enable.auto.commit"] = self.enable_auto_commit_bool
        #
        self.consumer = Consumer(self.consumer_config_dict)
        #
        self.subscribe()

    def __del__(self):
        self.close()

    #

    def subscribe(self):
        def on_assign(consumer, partitions):
            def set_offset(topicPartition):
                if topicPartition.topic in self.topic_str_start_offsets_dict_dict:
                    offsets = self.topic_str_start_offsets_dict_dict[topicPartition.topic]
                    if topicPartition.partition in offsets:
                        offset_int = offsets[topicPartition.partition]
                        topicPartition.offset = offset_int
                return topicPartition
            #
            if self.topic_str_start_offsets_dict_dict is not None:
                topicPartition_list = [set_offset(topicPartition) for topicPartition in partitions]
                consumer.assign(topicPartition_list)
        self.consumer.subscribe(self.topic_str_list, on_assign=on_assign)
        #
        return self.topic_str_list, self.group_str
    
    def unsubscribe(self):
        self.consumer.unsubscribe()
        #
        return self.topic_str_list, self.group_str

    def close(self):
        self.consumer.close()
        #
        return self.topic_str_list, self.group_str

    #

    def consume_impl(self, **kwargs):
        n_int = kwargs["n"] if "n" in kwargs and kwargs["n"] != ALL_MESSAGES else 1
        #
        message_list = self.consumer.consume(n_int, self.storage_obj.consume_timeout())
        #
        deserialized_message_dict_list = [{"topic": message.topic(), "headers": message.headers(), "partition": message.partition(), "offset": message.offset(), "timestamp": message.timestamp(), "key": self.deserialize(message.key(), self.topic_str_key_type_str_dict[message.topic()], topic_str=message.topic(), key_bool=True), "value": self.deserialize(message.value(), self.topic_str_value_type_str_dict[message.topic()], topic_str=message.topic(), key_bool=False)} for message in message_list]
        #
        return deserialized_message_dict_list

    #

    def commit(self, offsets=None, **kwargs):
        asynchronous_bool = kwargs["asynchronous"] if "asynchronous" in kwargs else False
        #
        if offsets is not None:
            str_or_int = list(offsets.keys())[0]
            if isinstance(str_or_int, str):
                topic_str_offsets_dict_dict = offsets
            elif isinstance(str_or_int, int):
                topic_str_offsets_dict_dict = {topic_str: offsets for topic_str in self.topic_str_list}
            #
            offsets_topicPartition_list = [TopicPartition(topic_str, partition_int, offset_int) for topic_str, offsets in topic_str_offsets_dict_dict.items() for partition_int, offset_int in offsets.items()]
            #
            commit_topicPartition_list = self.consumer.commit(offsets=offsets_topicPartition_list, asynchronous=asynchronous_bool)
            #
            topic_str_offsets_dict_dict = topicPartition_list_to_offsets_dict(commit_topicPartition_list)
        else:
            self.consumer.commit() # always returns None in this branch
            topic_str_offsets_dict_dict = {}
        #
        return topic_str_offsets_dict_dict

    def offsets(self, **kwargs):
        timeout_float = kwargs["timeout"] if "timeout" in kwargs else -1.0
        #
        assignment_topicPartition_list = self.consumer.assignment()
        committed_topicPartition_list = self.consumer.committed(assignment_topicPartition_list, timeout=timeout_float)
        #
        topic_str_offsets_dict_dict = topicPartition_list_to_offsets_dict(committed_topicPartition_list)
        #
        return topic_str_offsets_dict_dict

    def memberid(self):
        member_id_str = self.consumer.memberid()
        #
        return member_id_str 

#

def topicPartition_list_to_offsets_dict(topicPartition_list):
    offsets_dict = {}
    for topicPartition in topicPartition_list:
        topic_str = topicPartition.topic
        partition_int = topicPartition.partition
        offset_int = topicPartition.offset
        #
        if topic_str in offsets_dict:
            offsets = offsets_dict[topic_str]
        else:
            offsets = {}
        offsets[partition_int] = offset_int
        offsets_dict[topic_str] = offsets
    #
    return offsets_dict



from kafi.functional import Functional

# Constants

ALL_MESSAGES = -1

#

class AddOns(Functional):
    def compact(self, topic, n=ALL_MESSAGES, **kwargs):
        def foldl_function(acc, message_dict):
            key_hash_int_message_dict_dict = acc
            #
            key = message_dict["key"]
            value = message_dict["value"]
            #
            if key is not None:
                key_hash_int = hash(str(key))
                if value is None:
                    if key_hash_int in key_hash_int_message_dict_dict:
                        del key_hash_int_message_dict_dict[key_hash_int]
                else:
                    key_hash_int_message_dict_dict[key_hash_int] = message_dict
            #
            return key_hash_int_message_dict_dict
        #

        (key_hash_int_message_dict_dict, _) = self.foldl(topic, foldl_function, {}, n, **kwargs)
        #
        message_dict_list = list(key_hash_int_message_dict_dict.values())
        #
        return message_dict_list

    def compact_to(self, topic, target_storage, target_topic, n=ALL_MESSAGES, **kwargs):
        source_kwargs = self.get_source_kwargs(**kwargs)
        target_kwargs = self.get_target_kwargs(**kwargs)
        #
        message_dict_list = self.compact(topic, n, **source_kwargs)
        #
        target_producer = target_storage.producer(target_topic, **target_kwargs)
        key_bytes_list_value_bytes_list_tuple = target_producer.produce_list(message_dict_list, **target_kwargs)
        target_producer.close()
        #
        return key_bytes_list_value_bytes_list_tuple

    #

    def join_to(self, source_topic1, source_storage2, source_topic2, target_storage, target_topic, get_key_function1, get_key_function2, projection_function, join="left", n=ALL_MESSAGES, **kwargs):
        join_str = join
        #
        if join_str not in ["inner", "left", "right"]:
            raise Exception("Only \"inner\", \"left\" and \"right\" supported.")
        #
        def zip_foldl_to_function(acc, message_dict1, message_dict2):
            # print(message_dict1["value"])
            # print(message_dict2["value"])
            # print("===")
            (index_dict1, index_dict2) = acc
            #
            key1 = get_key_function1(message_dict1)
            key2 = get_key_function2(message_dict2)
            # DBSP: L join R = deltaL join deltaR + deltaL join R + L join deltaR
            out_message_dict_list = []
            # 1. deltaL join deltaR
            if key1 == key2:
                # Match in deltaL join deltaR.
                out_message_dict_list.append(projection_function(message_dict1, message_dict2))
            else:
                # 2. deltaL join R
                if key1 in index_dict2:
                    # Match in deltaL join R.
                    out_message_dict_list.append(projection_function(message_dict1, index_dict2[key1]))
                else:
                    # Could not find key1 in index_dict2.
                    # Only append to the output if the join type is "left"
                    if join_str == "left":
                        out_message_dict_list.append(message_dict1)
                # 3. L join deltaR
                if key2 in index_dict1:
                    # Match in L join deltaR
                    out_message_dict_list.append(projection_function(index_dict1[key2], message_dict2))
                else:
                    # Could not find key2 in index_dict1.
                    # Only append to the output if the join type is "right"
                    if join_str == "right":
                        out_message_dict_list.append(message_dict2)
            # Depending on the join type, persist:
            # * both sides (inner)
            # * the left side (left)
            # * the right side (right)
            if join_str == "inner":
                index_dict1[key1] = message_dict1
                index_dict2[key2] = message_dict2
            elif join_str == "left":
                index_dict1[key1] = message_dict1
            elif join_str == "right":
                index_dict1[key2] = message_dict2
            #
            return ((index_dict1, index_dict2), list(out_message_dict_list))

        return self.zip_foldl_to(source_topic1, source_storage2, source_topic2, target_storage, target_topic, zip_foldl_to_function, ({}, {}), n=n, **kwargs)

    #

    def recreate(self, topic, partitions=None, config={}, **kwargs):
        topic_str = topic
        #
        if self.exists(topic_str):
            if partitions is None:
                partitions_int = self.partitions(topic_str)[topic_str]
            else:
                partitions_int = partitions
            #
            old_config_dict = self.config(topic_str)[topic_str]
            config_dict = {}
            for key_str, value_str in old_config_dict.items():
                if key_str in config:
                    config_dict[key_str] = config[key_str]
                else:
                    config_dict[key_str] = value_str
            #
            self.delete(topic_str)
            #
            self.create(topic_str, partitions=partitions_int, config=config_dict, **kwargs)
        else:
            if partitions is None:
                partitions_int = 1
            else:
                partitions_int = partitions
            #
            self.create(topic_str, partitions=partitions_int, config=config, **kwargs)
        #
        return topic_str

    retouch = recreate



import glob
import os
import re

from piny import YamlLoader

from kafi.shell import Shell
from kafi.files import Files
from kafi.addons import AddOns
from kafi.schemaregistry import SchemaRegistry
from kafi.helpers import bytes_or_str_to_bytes, is_interactive

class Storage(Shell, Files, AddOns, SchemaRegistry):
    def __init__(self, dir_str, config_str, mandatory_section_str_list, optional_section_str_list):
        self.dir_str = dir_str
        self.config_str = config_str
        self.mandatory_section_str_list = mandatory_section_str_list
        self.optional_section_str_list = optional_section_str_list
        #
        self.config_dict = self.get_config_dict(config_str)
        #
        self.schema_registry_config_dict = self.config_dict["schema_registry"] if "schema_registry" in self.config_dict else {}
        #
        self.kafi_config_dict = self.config_dict["kafi"] if "kafi" in self.config_dict else self.config_dict["kash"] if "kash" in self.config_dict else {}
        #
        if "schema.registry.url" in self.schema_registry_config_dict:
            SchemaRegistry.__init__(self, self.schema_registry_config_dict)
        #
        if "progress.num.messages" not in self.kafi_config_dict:
            self.progress_num_messages(1000)
        else:
            self.progress_num_messages(int(self.kafi_config_dict["progress.num.messages"]))
        #
        if "consume.batch.size" not in self.kafi_config_dict:
            self.consume_batch_size(1000)
        else:
            self.consume_batch_size(int(self.kafi_config_dict["consume.batch.size"]))
        #
        if "produce.batch.size" not in self.kafi_config_dict:
            self.produce_batch_size(1000)
        else:
            self.produce_batch_size(int(self.kafi_config_dict["produce.batch.size"]))
        #
        if "verbose" not in self.kafi_config_dict:
            verbose_int = 1 if is_interactive() else 0
            self.verbose(verbose_int)
        else:
            self.verbose(int(self.kafi_config_dict["verbose"]))
        #
        if "auto.offset.reset" not in self.kafi_config_dict:
            self.auto_offset_reset("earliest")
        else:
            self.auto_offset_reset(str(self.kafi_config_dict["auto.offset.reset"]))
        #
        if "consumer.group.prefix" not in self.kafi_config_dict:
            self.consumer_group_prefix("")
        else:
            self.consumer_group_prefix(str(self.kafi_config_dict["consumer.group.prefix"]))
        #
        if "enable.auto.commit" not in self.kafi_config_dict:
            self.enable_auto_commit(False)
        else:
            self.enable_auto_commit(bool(self.kafi_config_dict["enable.auto.commit"]))
        #
        if "commit.after.processing" not in self.kafi_config_dict:
            self.commit_after_processing(True)
        else:
            self.commit_after_processing(bool(self.kafi_config_dict["commit.after.processing"]))
        #
        if "key.type" not in self.kafi_config_dict:
            self.key_type("str")
        else:
            self.key_type(str(self.kafi_config_dict["key.type"]))
        #
        if "value.type" not in self.kafi_config_dict:
            self.value_type("json")
        else:
            self.value_type(str(self.kafi_config_dict["value.type"]))

    #

    def progress_num_messages(self, new_value=None): # int
        return self.get_set_config("progress.num.messages", new_value)

    def consume_batch_size(self, new_value=None): # int
        return self.get_set_config("consume.batch.size", new_value)
    
    def produce_batch_size(self, new_value=None): # int
        return self.get_set_config("produce.batch.size", new_value)

    def verbose(self, new_value=None): # int
        return self.get_set_config("verbose", new_value)

    def auto_offset_reset(self, new_value=None): # str
        return self.get_set_config("auto.offset.reset", new_value)

    def consumer_group_prefix(self, new_value=None): # str
        return self.get_set_config("consumer.group.prefix", new_value)

    def enable_auto_commit(self, new_value=None): # bool
        return self.get_set_config("enable.auto.commit", new_value)

    def commit_after_processing(self, new_value=None): # bool
        return self.get_set_config("commit.after.processing", new_value)

    def key_type(self, new_value=None): # str
        return self.get_set_config("key.type", new_value)

    def value_type(self, new_value=None): # str
        return self.get_set_config("value.type", new_value)

    #

    def get_set_config(self, config_key_str, new_value=None, dict=None):
        dict = self.kafi_config_dict if dict is None else dict
        #
        if new_value is not None:
            dict[config_key_str] = new_value
        #
        return dict[config_key_str]

    #

    def get_config_dict(self, config_str):
        home_str = os.environ.get("KAFI_HOME")
        if not home_str:
            home_str = "."
        #
        config_dict = None
        configs_path_str_list = [f"{home_str}/configs/{self.dir_str}", f"{home_str}/configs", f"{home_str}"]
        for configs_path_str in configs_path_str_list:
            if os.path.exists(f"{configs_path_str}/{config_str}.yaml"):
                config_dict = YamlLoader(f"{configs_path_str}/{config_str}.yaml").load()
            elif os.path.exists(f"{configs_path_str}/{self.config_str}.yml"):
                config_dict = YamlLoader(f"{configs_path_str}/{config_str}.yml").load()
        if config_dict is None:
            raise Exception(f"No configuration file \"{config_str}.yaml\" or \"{config_str}.yml\" found in \"{configs_path_str_list}\" (hint: you can use KAFI_HOME environment variable to set the kafi home directory).")
        #
        for mandatory_section_str in self.mandatory_section_str_list:
            if mandatory_section_str not in config_dict:
                raise Exception(f"Connection configuration file \"{config_str}.yaml\" does not include a \"{mandatory_section_str}\" section.")
        #
        for optional_section_str in self.optional_section_str_list:
            if optional_section_str not in config_dict:
                config_dict[optional_section_str] = {}
        #
        return config_dict

    def configs(self, pattern="*", verbose=False):
        pattern_str = pattern
        verbose_bool = verbose
        #
        home_str = os.environ.get("KAFI_HOME")
        if not home_str:
            home_str = "."
        #
        configs_path_str = f"{home_str}/configs/{self.dir_str}"
        yaml_config_path_str_list = glob.glob(f"{configs_path_str}/{pattern_str}.yaml")
        yml_config_path_str_list = glob.glob(f"{configs_path_str}/{pattern_str}.yml")
        #
        yaml_config_str_list = [re.search(f"{configs_path_str}/(.*)\.yaml", yaml_config_path_str).group(1) for yaml_config_path_str in yaml_config_path_str_list if re.search(".*/(.*)\.yaml", yaml_config_path_str) is not None]
        yml_config_str_list = [re.search(".*/(.*)\.yml", yml_config_path_str).group(1) for yml_config_path_str in yml_config_path_str_list if re.search(".*/(.*)\.yml", yml_config_path_str) is not None]
        #
        config_str_list = yaml_config_str_list + yml_config_str_list
        #
        if verbose_bool:
            config_str_config_dict_dict = {config_str: self.get_config_dict(config_str) for config_str in config_str_list}
            return config_str_config_dict_dict
        else:
            config_str_list.sort()
            return config_str_list

    def is_headers_tuple_list(self, headers):
        return isinstance(headers, list) and len(headers) > 0 and all(isinstance(header_tuple, tuple) and len(header_tuple) == 2 and isinstance(header_tuple[0], str) for header_tuple in headers)


    def is_headers_dict(self, headers):
        return isinstance(headers, dict) and len(headers) > 0 and all(isinstance(header_key, str) for header_key in headers.keys())


    def is_headers(self, headers):
        return headers == None or self.is_headers_tuple_list(headers) or self.is_headers_dict(headers)

    def headers_to_headers_str_bytes_tuple_list(self, headers):
        if headers is None:
            headers_str_bytes_tuple_list = None
        elif self.is_headers_tuple_list(headers):
            headers_str_bytes_tuple_list = [(header_tuple[0], bytes_or_str_to_bytes(header_tuple[1])) for header_tuple in headers]
        elif self.is_headers_dict(headers):
            headers_str_bytes_tuple_list = [(header_key_str, bytes_or_str_to_bytes(header_value_str_or_bytes)) for header_key_str, header_value_str_or_bytes in headers.items()]
        else:
            raise Exception("Type error: Headers must either be a list of tuples of strings and bytes, or a dictionary of strings and bytes.")
        #
        return headers_str_bytes_tuple_list

    # Topics

    def topics(self, pattern=None, size=False, **kwargs):
        return self.admin.topics(pattern, size, **kwargs)
    
    ls = topics

    def l(self, pattern=None, size=True, **kwargs):
        return self.admin.topics(pattern=pattern, size=size, **kwargs)

    ll = l

    def exists(self, topic):
        topic_str = topic
        #
        return self.admin.topics(topic_str) != []

    #

    def watermarks(self, pattern, **kwargs):
        return self.admin.watermarks(pattern, **kwargs)

    def delete_records(self, pattern_or_offsets, **kwargs):
        return self.admin.delete_records(pattern_or_offsets, **kwargs)

    def lags(self, group_pattern, topic_pattern, **kwargs):
      group_offsets = self.admin.group_offsets(group_pattern)
      topic_str_partition_int_offsets_tuple_dict_dict = self.admin.watermarks(topic_pattern, **kwargs)
      #
      group_str_topic_str_lags_dict_dict_dict = {}
      for group_str, topic_str_group_offsets_dict_dict in group_offsets.items():
          group_str_topic_str_lags_dict_dict_dict[group_str] = {}
          for topic_str, group_offsets_dict in topic_str_group_offsets_dict_dict.items():
              group_str_topic_str_lags_dict_dict_dict[group_str][topic_str] = {partition_int: topic_str_partition_int_offsets_tuple_dict_dict[topic_str][partition_int][1] - group_offset_int for partition_int, group_offset_int in group_offsets_dict.items()}
      #
      return group_str_topic_str_lags_dict_dict_dict

    def config(self, pattern, config=None, **kwargs):
        return self.admin.config(pattern, config, **kwargs)

    def create(self, topic, partitions=1, **kwargs):
        return self.admin.create(topic, partitions, **kwargs)
    
    touch = create

    def delete(self, pattern, **kwargs):
        return self.admin.delete(pattern, **kwargs)

    rm = delete

    def offsets_for_times(self, pattern, partitions_timestamps, **kwargs):
        return self.admin.offsets_for_times(pattern, partitions_timestamps, **kwargs)

    def partitions(self, pattern, partitions=None, verbose=False, **kwargs):
        return self.admin.partitions(pattern, partitions, verbose, **kwargs)

    # Groups

    def groups(self, pattern="*", state_pattern="*", state=False):
        return self.admin.groups(pattern, state_pattern, state)
    
    def describe_groups(self, pattern="*", state_pattern="*"):
        return self.admin.describe_groups(pattern, state_pattern)
    
    def delete_groups(self, pattern, state_pattern="*"):
        return self.admin.delete_groups(pattern, state_pattern)
    
    rm_groups = delete_groups

    def group_offsets(self, pattern, group_offsets=None, state_pattern="*"):
        return self.admin.group_offsets(pattern, group_offsets, state_pattern)

    #

    def consumer(self, topic, **kwargs):
        consumer = self.get_consumer(topic, **kwargs)
        #
        return consumer
    
    def producer(self, topic, **kwargs):
        producer = self.get_producer(topic, **kwargs)
        #
        return producer

    # Helpers

    def get_key_value_type_tuple(self, **kwargs):
        # Default key and value types.
        key_type = self.key_type()
        value_type = self.value_type()
        #
        if "type" in kwargs:
            key_type = kwargs["type"]
            value_type = key_type
        #
        if "key_type" in kwargs:
            key_type = kwargs["key_type"]
        #
        if "value_type" in kwargs:
            value_type = kwargs["value_type"]
        #
        return (key_type, value_type)



import os
import importlib
import json
import sys
import tempfile

from confluent_kafka.schema_registry.avro import AvroDeserializer
from confluent_kafka.schema_registry.json_schema import JSONDeserializer
from confluent_kafka.schema_registry.protobuf import ProtobufDeserializer
from confluent_kafka.serialization import MessageField, SerializationContext
from google.protobuf.json_format import MessageToDict

from kafi.schemaregistry import SchemaRegistry

class Deserializer(SchemaRegistry):
    def __init__(self, schema_registry_config_dict):
        super().__init__(schema_registry_config_dict)

    def deserialize(self, payload_bytes, type_str, topic_str, key_bool):
        if type_str.lower() == "bytes":
            deserialized_payload = self.bytes_to_bytes(payload_bytes)
        elif type_str.lower() == "str":
            deserialized_payload = self.bytes_to_str(payload_bytes)
        elif type_str.lower() == "json":
            deserialized_payload = self.bytes_to_dict(payload_bytes)
        elif type_str.lower() in ["protobuf", "pb"]:
            deserialized_payload = self.bytes_protobuf_to_dict(payload_bytes, topic_str, key_bool)
        elif type_str.lower() == "avro":
            deserialized_payload = self.bytes_avro_to_dict(payload_bytes, topic_str, key_bool)
        elif type_str.lower() in ["jsonschema", "json_sr"]:
            deserialized_payload = self.bytes_jsonschema_to_dict(payload_bytes, topic_str, key_bool)
        else:
            raise Exception("Only \"str\", \"bytes\", \"json\", \"protobuf\" (\"pb\"), \"avro\" and \"jsonschema\" (\"json_sr\") supported.")
        #
        return deserialized_payload

    def bytes_to_str(self, bytes):
        if bytes:
            return bytes.decode("utf-8")
        else:
            return bytes

    def bytes_to_bytes(self, bytes):
        return bytes

    def bytes_to_dict(self, bytes):
        if bytes is None:
            return None
        #
        return json.loads(bytes)

    def bytes_protobuf_to_dict(self, bytes, topic_str, key_bool):
        if bytes is None:
            return None
        #
        schema_id_int = int.from_bytes(bytes[1:5], "big")
        if schema_id_int in self.schema_id_int_generalizedProtocolMessageType_protobuf_schema_str_tuple_dict:
            generalizedProtocolMessageType, protobuf_schema_str = self.schema_id_int_generalizedProtocolMessageType_protobuf_schema_str_tuple_dict[schema_id_int]
        else:
            generalizedProtocolMessageType, protobuf_schema_str = self.schema_id_int_to_generalizedProtocolMessageType_protobuf_schema_str_tuple(schema_id_int)
            self.schema_id_int_generalizedProtocolMessageType_protobuf_schema_str_tuple_dict[schema_id_int] = (generalizedProtocolMessageType, protobuf_schema_str)
        #
        protobufDeserializer = ProtobufDeserializer(generalizedProtocolMessageType, {"use.deprecated.format": False})
        serializationContext = SerializationContext(topic_str, MessageField.KEY if key_bool else MessageField.VALUE)
        protobuf_message = protobufDeserializer(bytes, serializationContext)
        dict = MessageToDict(protobuf_message)
        return dict

    def bytes_avro_to_dict(self, bytes, topic_str, key_bool):
        if bytes is None:
            return None
        #
        schema_id_int = int.from_bytes(bytes[1:5], "big")
        schema_dict = self.get_schema(schema_id_int)
        schema_str = schema_dict["schema_str"]
        #
        avroDeserializer = AvroDeserializer(self.schemaRegistryClient, schema_str)
        serializationContext = SerializationContext(topic_str, MessageField.KEY if key_bool else MessageField.VALUE)
        dict = avroDeserializer(bytes, serializationContext)
        return dict

    def bytes_jsonschema_to_dict(self, bytes, topic_str, key_bool):
        if bytes is None:
            return None
        #
        schema_id_int = int.from_bytes(bytes[1:5], "big")
        schema_dict = self.get_schema(schema_id_int)
        schema_str = schema_dict["schema_str"]
        #
        jsonDeserializer = JSONDeserializer(schema_str)
        serializationContext = SerializationContext(topic_str, MessageField.KEY if key_bool else MessageField.VALUE)
        dict = jsonDeserializer(bytes, serializationContext)
        return dict

    # Helpers

    def schema_id_int_to_generalizedProtocolMessageType_protobuf_schema_str_tuple(self, schema_id_int):
        schema_dict = self.get_schema(schema_id_int)
        schema_str = schema_dict["schema_str"]
        #
        generalizedProtocolMessageType = self.schema_id_int_and_schema_str_to_generalizedProtocolMessageType(schema_id_int, schema_str)
        #
        return generalizedProtocolMessageType, schema_str

    def schema_id_int_and_schema_str_to_generalizedProtocolMessageType(self, schema_id_int, schema_str):
        path_str = f"/{tempfile.gettempdir()}/kafi/clusters/{self.storage_obj.config_str}"
        os.makedirs(path_str, exist_ok=True)
        file_str = f"schema_{schema_id_int}.proto"
        file_path_str = f"{path_str}/{file_str}"
        with open(file_path_str, "w") as textIOWrapper:
            textIOWrapper.write(schema_str)
        #
        import grpc_tools.protoc
        grpc_tools.protoc.main(["protoc", f"-I{path_str}", f"--python_out={path_str}", f"{file_str}"])
        #
        sys.path.insert(1, path_str)
        schema_module = importlib.import_module(f"schema_{schema_id_int}_pb2")
        schema_name_str = list(schema_module.DESCRIPTOR.message_types_by_name.keys())[0]
        generalizedProtocolMessageType = getattr(schema_module, schema_name_str)
        return generalizedProtocolMessageType



import ast
import os

from kafi.storage_consumer import StorageConsumer

# Constants

ALL_MESSAGES = -1
OFFSET_INVALID = -1001

#

class FSConsumer(StorageConsumer):
    def __init__(self, fs_obj, *topics, **kwargs):
        super().__init__(fs_obj, *topics, **kwargs)
        #  required for commit() without offsets.
        # Initialize self.next_topic_str_offsets_dict_dict
        self.next_topic_str_offsets_dict_dict = {topic_str: {partition_int: OFFSET_INVALID for partition_int in range(self.storage_obj.admin.get_partitions(topic_str))} for topic_str in self.topic_str_list}
        # Initialize/update group dict file.
        group_dict = self.storage_obj.admin.get_group_dict(self.group_str)
        if group_dict == {}:
            group_dict = {"offsets": {}, "state": "stable"}
        for topic_str in self.topic_str_list:
            if topic_str not in group_dict["offsets"]:
                # if there are no offsets for the topic yet, use the defaults.
                group_dict["offsets"][topic_str] = self.next_topic_str_offsets_dict_dict[topic_str]
        self.storage_obj.admin.set_group_dict(self.group_str, group_dict)
            
    #

    def close(self):
        new_group_dict = {"state": "empty"}
        self.storage_obj.admin.set_group_dict(self.group_str, new_group_dict)
    
    #
  
    def foldl(self, foldl_function, initial_acc, n=ALL_MESSAGES, commit_after_processing=None, **kwargs):
        n_int = n
        #
        commit_after_processing_bool = self.storage_obj.commit_after_processing() if commit_after_processing is None else commit_after_processing
        #
        auto_offset_reset_str = self.consumer_config_dict["auto.offset.reset"]
        #
        message_counter_int = 0
        acc = initial_acc
        rel_file_str_list = []
        for topic_str in self.topic_str_list:
            partitions_int = self.storage_obj.admin.get_partitions(topic_str)
            #
            # Get start offsets.
            start_offsets_dict = self.topic_str_start_offsets_dict_dict[topic_str] if self.topic_str_start_offsets_dict_dict is not None and topic_str in self.topic_str_start_offsets_dict_dict else None
            #
            if start_offsets_dict is None:
                # If we got no start offsets, try to get the start offsets from the group.
                group_dict = self.storage_obj.admin.get_group_dict(self.group_str)
                group_offsets_dict = group_dict["offsets"][topic_str]
                #
                if any(group_offsets_dict[partition_int] == OFFSET_INVALID for partition_int in range(partitions_int)):
                    # If any of the partitions still does not have a committed offset yet...
                    if auto_offset_reset_str.lower() == "latest":
                        # If auto.offset.reset == latest, start offsets = last offsets.
                        partition_int_offset_tuple_dict = self.watermarks(topic_str, **kwargs)[topic_str]
                        auto_offset_reset_offsets_dict = {partition_int: offset_tuple[1] for partition_int, offset_tuple in partition_int_offset_tuple_dict.items()}
                    elif auto_offset_reset_str.lower() == "earliest":
                        # If auto.offset.reset == earliest, start offsets = 0
                        auto_offset_reset_offsets_dict = {partition_int: 0 for partition_int in range(partitions_int)}
                    else:
                        raise Exception("Only \"earliest\" and \"latest\" supported for \"auto.offset.reset\".")
                    # ...and for each partition still not having a committed offset yet, use the auto.offset.reset one.
                    start_offsets_dict = {}
                    for partition_int, offset_int in group_offsets_dict.items():
                        if offset_int == OFFSET_INVALID:
                            start_offsets_dict[partition_int] = auto_offset_reset_offsets_dict[partition_int]
                        else:
                            start_offsets_dict[partition_int] = group_offsets_dict[partition_int]
                else:
                    start_offsets_dict = group_offsets_dict
            #
            # Get partition files for all partitions.
            partition_int_rel_file_str_list_dict = self.storage_obj.admin.get_partition_files(topic_str)
            #
            # Get first partition files for all partitions.
            partition_int_first_partition_rel_file_str_dict = {partition_int: self.storage_obj.admin.find_partition_file_str_by_offset(topic_str, partition_int, offset_int) for partition_int, offset_int in start_offsets_dict.items()}
            #
            # Filter out partitions not corresponding to any filee listed by get_partition_files() above.
            partition_int_first_partition_rel_file_str_dict = {partition_int: first_partition_rel_file_str for partition_int, first_partition_rel_file_str in partition_int_first_partition_rel_file_str_dict.items() if first_partition_rel_file_str is not None}
            #
            # Get all partition files to be consumed for all partitions.
            partition_int_to_be_consume_rel_file_str_list_dict = {partition_int: [rel_file_str for rel_file_str in rel_file_str_list if partition_int in partition_int_first_partition_rel_file_str_dict and rel_file_str >= partition_int_first_partition_rel_file_str_dict[partition_int]] for partition_int, rel_file_str_list in partition_int_rel_file_str_list_dict.items()}
            #
            # Create list of partition files to read.
            file_counter_int = 0
            max_num_files_int = max([len(to_be_consume_rel_file_str_list) for to_be_consume_rel_file_str_list in partition_int_to_be_consume_rel_file_str_list_dict.values()])
            #
            for file_counter_int in range(max_num_files_int):
                for partition_int in range(partitions_int):
                    if partition_int in partition_int_to_be_consume_rel_file_str_list_dict:
                        if len(partition_int_to_be_consume_rel_file_str_list_dict[partition_int]) > file_counter_int:
                            rel_file_str_list.append(partition_int_to_be_consume_rel_file_str_list_dict[partition_int][file_counter_int])
            #
            abs_topic_dir_str = self.storage_obj.admin.get_topic_abs_path_str(topic_str)
            for rel_file_str in rel_file_str_list:
                messages_bytes = self.storage_obj.admin.read_bytes(os.path.join(abs_topic_dir_str, "partitions", rel_file_str))
                #
                message_bytes_list = messages_bytes.split(b"\n")[:-1]
                #
                for message_bytes in message_bytes_list:
                    message_dict = ast.literal_eval(message_bytes.decode("utf-8"))
                    #
                    message_dict["key"] = self.deserialize(message_dict["key"], self.topic_str_key_type_str_dict[message_dict["topic"]], topic_str=topic_str, key_bool=True)
                    #
                    message_dict["value"] = self.deserialize(message_dict["value"], self.topic_str_value_type_str_dict[message_dict["topic"]], topic_str=topic_str, key_bool=False)
                    #
                    partition_int = message_dict["partition"]
                    self.next_topic_str_offsets_dict_dict[topic_str][partition_int] = message_dict["offset"] + 1
                    if self.enable_auto_commit_bool:
                        # Commit immediately after reading the message if enable.auto.commit == True
                        self.commit()
                    #
                    if message_dict["offset"] >= start_offsets_dict[partition_int]:
                        acc = foldl_function(acc, message_dict)
                        #
                        message_counter_int += 1
                        #
                        if not self.enable_auto_commit_bool and commit_after_processing_bool:
                            # Only commit once the message has been processed if enable.auto.commit == False and commit.after.processing == True
                            self.commit()
                    #
                    if n_int != ALL_MESSAGES:
                        if message_counter_int >= n_int:
                            return acc
        #
        return acc

    #

    def consume(self, n=ALL_MESSAGES):
        def foldl_function(message_dict_list, message_dict):
            message_dict_list.append(message_dict)
            #
            return message_dict_list
        #
        return self.foldl(foldl_function, [], n)

    def offsets(self):
        group_str_topic_str_offsets_dict_dict_dict = self.storage_obj.admin.group_offsets(self.group_str)
        #
        if self.group_str not in group_str_topic_str_offsets_dict_dict_dict:
            topic_str_offsets_dict_dict = {topic_str: {} for topic_str in self.topic_str_list}
        else:
            topic_str_offsets_dict_dict = {topic_str: offsets_dict for topic_str, offsets_dict in group_str_topic_str_offsets_dict_dict_dict[self.group_str].items() if topic_str in self.topic_str_list}
        #
        return topic_str_offsets_dict_dict

    def commit(self, offsets=None):
        if offsets is None:
            new_group_dict = {"offsets": self.next_topic_str_offsets_dict_dict}
            #
            topic_str_offsets_dict_dict = {topic_str: self.next_topic_str_offsets_dict_dict[topic_str] for topic_str in self.topic_str_list}
        else:
            str_or_int = list(offsets.keys())[0]
            if isinstance(str_or_int, str):
                topic_str_offsets_dict_dict = offsets
            elif isinstance(str_or_int, int):
                topic_str_offsets_dict_dict = {topic_str: offsets for topic_str in self.topic_str_list}
            #
            new_group_dict = {"offsets": topic_str_offsets_dict_dict}
        #
        self.storage_obj.admin.set_group_dict(self.group_str, new_group_dict)
        #
        return topic_str_offsets_dict_dict



from kafi.fs.fs_producer import FSProducer


class AzureBlobProducer(FSProducer):
    def __init__(self, azureblob_obj, file, **kwargs):
        super().__init__(azureblob_obj, file, **kwargs)

    #

    def close(self):
        return self.topic_str



from kafi.fs.fs import FS
from kafi.fs.azureblob.azureblob_admin import AzureBlobAdmin
from kafi.fs.azureblob.azureblob_consumer import AzureBlobConsumer
from kafi.fs.azureblob.azureblob_producer import AzureBlobProducer

#

class AzureBlob(FS):
    def __init__(self, config_str):
        super().__init__("azureblobs", config_str, ["azure_blob"], ["schema_registry"])
    
    #

    def get_admin(self):
        admin = AzureBlobAdmin(self)
        #
        return admin

    #

    def get_consumer(self, file, **kwargs):
        consumer = AzureBlobConsumer(self, file, **kwargs)
        #
        return consumer

    #

    def get_producer(self, file, **kwargs):
        producer = AzureBlobProducer(self, file, **kwargs)
        #
        return producer



from kafi.kafka.cluster.cluster_admin import ClusterAdmin
from kafi.kafka.cluster.cluster_consumer import ClusterConsumer
from kafi.kafka.cluster.cluster_producer import ClusterProducer
from kafi.kafka.kafka import Kafka
from kafi.helpers import is_interactive

# Cluster class

class Cluster(Kafka):
    def __init__(self, config_str):
        super().__init__("clusters", config_str, ["kafka"], ["schema_registry"])
        #
        # If not set already in the config, set librdkafka log level to:
        # * 3 ("Error") if interactive (=in the interpreter)
        # * 6 ("Notice", default for librdkafka) otherwise
        # See https://github.com/confluentinc/librdkafka/blob/master/CONFIGURATION.md
        # and https://en.wikipedia.org/wiki/Syslog
        if "log_level" not in self.kafka_config_dict:
            if is_interactive():
                self.kafka_config_dict["log_level"] = 3
            else:
                self.kafka_config_dict["log_level"] = 6
        #
        self.admin = self.get_admin()

    #

    def get_admin(self):
        admin = ClusterAdmin(self)
        #
        return admin

    #
    def get_consumer(self, topics, **kwargs):
        consumer = ClusterConsumer(self, topics, **kwargs)
        #
        return consumer

    #

    def get_producer(self, topic, **kwargs):
        producer = ClusterProducer(self, topic, **kwargs)
        #
        return producer



import os

from kafi.fs.fs_admin import FSAdmin

#

class AzureBlobAdmin(FSAdmin):
    def __init__(self, azureblob_obj):
        from azure.storage.blob import BlobServiceClient
        #

        super().__init__(azureblob_obj)
        #
        blobServiceClient = BlobServiceClient.from_connection_string(azureblob_obj.azure_blob_config_dict["connection.string"])
        self.containerClient = blobServiceClient.get_container_client(azureblob_obj.container_name())

    # Topics/Files

    def list_dirs(self, abs_path_dir_str):
        blob_str_itemPaged = self.containerClient.list_blob_names(name_starts_with=abs_path_dir_str)
        rel_dir_str_set = set([os.path.basename(os.path.dirname(blob_str)) for blob_str in blob_str_itemPaged])
        #
        rel_dir_str_list = list(rel_dir_str_set)
        rel_dir_str_list.sort()
        #
        return rel_dir_str_list

    def list_files(self, abs_path_dir_str):
        blob_str_itemPaged = self.containerClient.list_blob_names(name_starts_with=abs_path_dir_str)
        rel_file_str_set = set([os.path.relpath(blob_str, abs_path_dir_str) for blob_str in blob_str_itemPaged])
        #
        rel_file_str_list = list(rel_file_str_set)
        rel_file_str_list.sort()
        #
        return rel_file_str_list


    def delete_file(self, abs_path_file_str):
        self.containerClient.delete_blob(abs_path_file_str)

    def delete_dir(self, _):
        pass

    def exists_file(self, abs_path_file_str):
        from azure.storage.blob import BlobClient
        #

        blobClient = BlobClient.from_connection_string(self.storage_obj.azure_blob_config_dict["connection.string"], container_name=self.storage_obj.azure_blob_config_dict["container.name"], blob_name=abs_path_file_str)
        #
        return blobClient.exists()

    # Metadata
    
    def read_str(self, abs_path_file_str):
        from azure.storage.blob import BlobClient
        #

        blobClient = BlobClient.from_connection_string(conn_str=self.storage_obj.azure_blob_config_dict["connection.string"], container_name=self.storage_obj.azure_blob_config_dict["container.name"], blob_name=abs_path_file_str)
        #
        storageStreamDownloader = blobClient.download_blob()
        blob_bytes = storageStreamDownloader.read()
        #
        blob_str = blob_bytes.decode("utf-8")
        #
        return blob_str

    def write_str(self, abs_path_file_str, data_str):
        from azure.storage.blob import BlobClient
        #

        blobClient = BlobClient.from_connection_string(conn_str=self.storage_obj.azure_blob_config_dict["connection.string"], container_name=self.storage_obj.azure_blob_config_dict["container.name"], blob_name=abs_path_file_str)
        #
        data_bytes = data_str.encode("utf-8")
        #
        blobClient.upload_blob(data_bytes, overwrite=True)        

    #

    def read_bytes(self, abs_path_file_str):
        from azure.storage.blob import BlobClient
        #

        blobClient = BlobClient.from_connection_string(conn_str=self.storage_obj.azure_blob_config_dict["connection.string"], container_name=self.storage_obj.azure_blob_config_dict["container.name"], blob_name=abs_path_file_str)
        #
        storageStreamDownloader = blobClient.download_blob()
        blob_bytes = storageStreamDownloader.read()
        #
        return blob_bytes

    def write_bytes(self, abs_path_file_str, data_bytes):
        from azure.storage.blob import BlobClient
        #

        blobClient = BlobClient.from_connection_string(conn_str=self.storage_obj.azure_blob_config_dict["connection.string"], container_name=self.storage_obj.azure_blob_config_dict["container.name"], blob_name=abs_path_file_str)
        #
        blobClient.upload_blob(data_bytes)



import os

from kafi.fs.fs_admin import FSAdmin

#

class LocalAdmin(FSAdmin):
    def __init__(self, local_obj, **kwargs):
        super().__init__(local_obj, **kwargs)

    # Topics/Files

    def list_dirs(self, abs_path_dir_str):
        collected_rel_dir_str_list = []
        for abs_root_dir_str, rel_dir_str_list, _ in os.walk(abs_path_dir_str):
            rel_root_dir_str = os.path.relpath(abs_root_dir_str, abs_path_dir_str)
            #
            collected_rel_dir_str_list += [rel_dir_str if rel_root_dir_str == "." else os.path.join(rel_root_dir_str, rel_dir_str) for rel_dir_str in rel_dir_str_list]
        #
        collected_rel_dir_str_list.sort()
        #
        return collected_rel_dir_str_list

    def list_files(self, abs_path_dir_str):
        collected_rel_file_str_list = []
        for abs_root_dir_str, _, rel_file_str_list in os.walk(abs_path_dir_str):
            rel_root_dir_str = os.path.relpath(abs_root_dir_str, abs_path_dir_str)
            #
            collected_rel_file_str_list += [rel_file_str if rel_root_dir_str == "." else os.path.join(rel_root_dir_str, rel_file_str) for rel_file_str in rel_file_str_list]
        #
        collected_rel_file_str_list.sort()
        #
        return collected_rel_file_str_list

    def delete_file(self, abs_path_file_str):
        os.remove(abs_path_file_str)

    def delete_dir(self, abs_path_dir_str):
        os.rmdir(abs_path_dir_str)

    def exists_file(self, abs_path_file_str):
        return os.path.exists(abs_path_file_str)

    # Metadata
    
    def read_str(self, abs_path_file_str):
        str = None
        #
        if os.path.exists(abs_path_file_str):
            with open(abs_path_file_str, "r") as bufferedReader:
                str = bufferedReader.read()
        #
        return str

    def write_str(self, abs_path_file_str, data_str):
        os.makedirs(os.path.dirname(abs_path_file_str), exist_ok=True)
        #
        with open(abs_path_file_str, "w") as bufferedWriter:
            bufferedWriter.write(data_str)

    #

    def read_bytes(self, abs_path_file_str):
        with open(abs_path_file_str, "rb") as bufferedReader:
            bytes = bufferedReader.read()
        #
        return bytes

    def write_bytes(self, abs_path_file_str, data_bytes):
        os.makedirs(os.path.dirname(abs_path_file_str), exist_ok=True)
        #
        with open(abs_path_file_str, "wb") as bufferedWriter:
            bufferedWriter.write(data_bytes)



from kafi.fs.fs_producer import FSProducer

class S3Producer(FSProducer):
    def __init__(self, s3_obj, file, **kwargs):
        super().__init__(s3_obj, file, **kwargs)

    #

    def close(self):
        return self.topic_str




from kafi.storage_producer import StorageProducer

class KafkaProducer(StorageProducer):
    def __init__(self, kafka_obj, topic, **kwargs):
        super().__init__(kafka_obj, topic, **kwargs)

    #

    def write(self, value, **kwargs):
        return self.produce(value, **kwargs)



import os
import sys

if os.path.basename(os.getcwd()) == "test":
    sys.path.insert(1, "..")
else:
    sys.path.insert(1, ".")

from test.test_single_storage_base import TestSingleStorageBase
from kafi.kafka.restproxy.restproxy import RestProxy
from kafi.kafka.cluster.cluster import Cluster

class Test(TestSingleStorageBase):
    def setUp(self):
        super().setUp()
        self.principal_str = None


    def tearDown(self):
        super().tearDown()

    #

    def get_storage(self):
        r = RestProxy("local")
        #
        return r

    def get_cluster(self):
        c = Cluster("local")
        #
        return c

    def is_ccloud(self):
        r = self.get_storage()
        #
        return "confluent.cloud" in r.rest_proxy_config_dict["rest.proxy.url"]
    
    def test_delete_groups(self):
        pass

    def test_set_group_offsets(self):
        pass

    def test_lags(self):
        pass

    def test_offsets_for_times(self):
        pass

    def test_consume_from_offsets(self):
        pass

    def test_error_handling(self):
        pass

    def test_compact(self):
        pass



import os
import random
import sys
import tempfile
import unittest
import warnings

if os.path.basename(os.getcwd()) == "test":
    sys.path.insert(1, "..")
else:
    sys.path.insert(1, ".")

from kafi.kafi import *
from kafi.helpers import *

#

config_str = "local"

class Test(unittest.TestCase):
    def setUp(self):
        warnings.filterwarnings(action="ignore", message="unclosed", category=ResourceWarning)
        #
        # https://simon-aubury.medium.com/kafka-with-avro-vs-kafka-with-protobuf-vs-kafka-with-json-schema-667494cbb2af
        self.snack_str_list = ['{"name": "cookie", "calories": 500.0, "colour": "brown"}', '{"name": "cake", "calories": 260.0, "colour": "white"}', '{"name": "timtam", "calories": 80.0, "colour": "chocolate"}']
        self.snack_bytes_list = [bytes(snack_str, encoding="utf-8") for snack_str in self.snack_str_list]
        self.snack_dict_list = [json.loads(snack_str) for snack_str in self.snack_str_list]
        #
        self.snack_ish_dict_list = []
        for snack_dict in self.snack_dict_list:
            snack_dict1 = snack_dict.copy()
            snack_dict1["colour"] += "ish"
            self.snack_ish_dict_list.append(snack_dict1)
        #
        self.headers_str_bytes_tuple_list = [("header_field1", b"header_value1"), ("header_field2", b"header_value2")]
        self.headers_str_bytes_dict = {"header_field1": b"header_value1", "header_field2": b"header_value2"}
        self.headers_str_str_tuple_list = [("header_field1", "header_value1"), ("header_field2", "header_value2")]
        self.headers_str_str_dict = {"header_field1": "header_value1", "header_field2": "header_value2"}
        #
        self.avro_key_schema_str = '{ "type": "record", "name": "mykeyrecord", "fields": [{"name": "key",  "type": "string" }] }'
        self.avro_value_schema_str = '{ "type": "record", "name": "myvaluerecord", "fields": [{"name": "name",  "type": "string" }, {"name": "calories", "type": "float" }, {"name": "colour", "type": "string" }] }'
        self.protobuf_key_schema_str = 'message SnackKey { required string key = 1; }'
        self.protobuf_value_schema_str = 'message SnackValue { required string name = 1; required float calories = 2; optional string colour = 3; }'
        self.jsonschema_key_schema_str = '{ "title": "abckey", "definitions" : { "record:myrecordkey" : { "type" : "object", "required" : [ "key" ], "additionalProperties" : false, "properties" : { "key" : {"type" : "string"} } } }, "$ref" : "#/definitions/record:myrecordkey" }'
        self.jsonschema_value_schema_str = '{ "title": "abcvalue", "definitions" : { "record:myrecordvalue" : { "type" : "object", "required" : [ "name", "calories", "colour" ], "additionalProperties" : false, "properties" : { "name" : {"type" : "string"}, "calories" : {"type" : "number"}, "colour" : {"type" : "string"} } } }, "$ref" : "#/definitions/record:myrecordvalue" }'
        #
        self.key_schema_str_list = [self.avro_key_schema_str, self.protobuf_key_schema_str, self.jsonschema_key_schema_str]
        self.value_schema_str_list = [self.avro_value_schema_str, self.protobuf_value_schema_str, self.jsonschema_value_schema_str]
        #
        self.storage_str_topic_str_list_dict = {"Cluster": [], "RestProxy": [], "AzureBlob": [], "Local": [], "S3": []}
        self.storage_str_group_str_list_dict = {"Cluster": [], "RestProxy": [], "AzureBlob": [], "Local": [], "S3": []}
        #
        self.azureblob_s3_path_str = "test"
        self.local_path_str = f"{tempfile.gettempdir()}/kafi/test/local"
        os.makedirs(self.local_path_str, exist_ok=True)
        #
        print("Test:", self._testMethodName)

    def tearDown(self):
        c = Cluster("local")
        for group_str in self.storage_str_group_str_list_dict["Cluster"]:
            c.delete_groups(group_str)
        for group_str in self.storage_str_group_str_list_dict["RestProxy"]:
            c.delete_groups(group_str)
        for topic_str in self.storage_str_topic_str_list_dict["Cluster"]:
            c.delete(topic_str)
        #
        r = RestProxy("local")
        for topic_str in self.storage_str_topic_str_list_dict["RestProxy"]:
            r.delete(topic_str)
        #
        a = self.get_azureblob()
        for group_str in self.storage_str_group_str_list_dict["AzureBlob"]:
#            a.delete_groups(group_str)
            pass
        for topic_str in self.storage_str_topic_str_list_dict["AzureBlob"]:
            a.delete(topic_str)
        #
        l = self.get_local()
        for group_str in self.storage_str_group_str_list_dict["Local"]:
#            l.delete_groups(group_str)
            pass
        for topic_str in self.storage_str_topic_str_list_dict["Local"]:
            l.delete(topic_str)
        #
        s = self.get_s3()
        for group_str in self.storage_str_group_str_list_dict["S3"]:
#            s.delete_groups(group_str)
            pass
        for topic_str in self.storage_str_topic_str_list_dict["S3"]:
            s.delete(topic_str)

    def create_test_topic_name(self, storage_obj):
        topic_str = f"test_topic_{get_millis()}"
        while True:
            #
            storage_str = storage_obj.__class__.__name__
            if topic_str not in self.storage_str_topic_str_list_dict[storage_str]:
                self.storage_str_topic_str_list_dict[storage_str].append(topic_str)
                break
        #
        return topic_str

    def create_test_group_name(self, storage_obj):
        while True:
            group_str = f"test_group_{get_millis()}"
            #
            storage_str = storage_obj.__class__.__name__
            if group_str not in self.storage_str_group_str_list_dict[storage_str]:
                self.storage_str_group_str_list_dict[storage_str].append(group_str)
                break
        #
        return group_str

    def get_azureblob(self):
        a = AzureBlob(config_str)
        a.root_dir(self.azureblob_s3_path_str)
        return a

    def get_local(self):
        l = Local(config_str)
        l.root_dir(self.local_path_str)
        return l

    def get_s3(self):
        s = S3(config_str)
        s.root_dir(self.azureblob_s3_path_str)
        return s

    # Cp from fs storage to fs storage

    def test_cp_azureblob_to_local(self):
        a = self.get_azureblob()
        l = self.get_local()
        #
        test_cp(self, a, l)

    def test_cp_azureblob_to_s3(self):
        a = self.get_azureblob()
        s = self.get_s3()
        #
        test_cp(self, a, s)

    def test_cp_local_to_azureblob(self):
        l = self.get_local()
        a = self.get_azureblob()
        #
        test_cp(self, l, a)

    def test_cp_local_to_s3(self):
        l = self.get_local()
        s = self.get_s3()
        #
        test_cp(self, l, s)

    def test_cp_s3_to_azureblob(self):
        s = self.get_s3()
        a = self.get_azureblob()
        #
        test_cp(self, s, a)

    def test_cp_s3_to_local(self):
        s = self.get_s3()
        l = self.get_local()
        #
        test_cp(self, s, l)

    # Cp from kafka storage to fs storage

    def test_cp_cluster_to_azureblob(self):
        c = Cluster("local")
        a = self.get_azureblob()
        #
        test_cp(self, c, a)

    def test_cp_cluster_to_local(self):
        c = Cluster("local")
        l = self.get_local()
        #
        test_cp(self, c, l)

    def test_cp_cluster_to_s3(self):
        c = Cluster("local")
        s = self.get_s3()
        #
        test_cp(self, c, s)

    def test_cp_restproxy_to_azureblob(self):
        r = RestProxy("local")
        a = self.get_azureblob()
        #
        test_cp(self, r, a)

    def test_cp_restproxy_to_local(self):
        r = RestProxy("local")
        l = self.get_local()
        #
        test_cp(self, r, l)

    def test_cp_restproxy_to_s3(self):
        r = RestProxy("local")
        s = self.get_s3()
        #
        test_cp(self, r, s)

    # Cp from fs storage to kafka storage

    def test_cp_azureblob_to_cluster(self):
        a = self.get_azureblob()
        c = Cluster("local")
        #
        test_cp(self, a, c)

    def test_cp_local_to_cluster(self):
        l = self.get_local()
        c = Cluster("local")
        #
        test_cp(self, l, c)

    def test_cp_s3_to_cluster(self):
        s = self.get_s3()
        c = Cluster("local")
        #
        test_cp(self, s, c)

    def test_cp_azureblob_to_restproxy(self):
        a = self.get_azureblob()
        r = RestProxy("local")
        #
        test_cp(self, a, r)

    def test_cp_local_to_restproxy(self):
        l = self.get_local()
        r = RestProxy("local")
        #
        test_cp(self, l, r)

    def test_cp_s3_to_restproxy(self):
        s = self.get_s3()
        r = RestProxy("local")
        #
        test_cp(self, s, r)

    #

    def test_cp_cluster_to_restproxy(self):
        c = Cluster("local")
        r = RestProxy("local")
        #
        test_cp(self, c, r)

    def test_cp_restproxy_to_cluster(self):
        r = RestProxy("local")
        c = Cluster("local")
        test_cp(self, r, c)

    # Diff AzureBlob and X

    def test_diff_azureblob_local(self):
        a = self.get_azureblob()
        l = self.get_local()
        #        
        test_diff(self, a, l)

    def test_diff_azureblob_s3(self):
        a = self.get_azureblob()
        s = self.get_s3()
        #        
        test_diff(self, a, s)

    def test_diff_azureblob_cluster(self):
        a = self.get_azureblob()
        c = Cluster("local")
        #        
        test_diff(self, a, c)

    def test_diff_azureblob_restproxy(self):
        a = self.get_azureblob()
        r = RestProxy("local")
        #        
        test_diff(self, a, r)

    # Diff Local and X

    def test_diff_local_azureblob(self):
        l = self.get_local()
        a = self.get_azureblob()
        #        
        test_diff(self, l, a)

    def test_diff_local_s3(self):
        l = self.get_local()
        s = self.get_s3()
        #        
        test_diff(self, l, s)

    def test_diff_local_cluster(self):
        l = self.get_local()
        c = Cluster("local")
        #        
        test_diff(self, l, c)

    def test_diff_local_restproxy(self):
        l = self.get_local()
        r = RestProxy("local")
        #        
        test_diff(self, l, r)

    # Diff S3 and X

    def test_diff_s3_azureblob(self):
        s = self.get_s3()
        a = self.get_azureblob()
        #        
        test_diff(self, s, s)

    def test_diff_s3_local(self):
        s = self.get_s3()
        l = self.get_local()
        #        
        test_diff(self, s, l)

    def test_diff_s3_cluster(self):
        s = self.get_s3()
        c = Cluster("local")
        #        
        test_diff(self, s, c)

    def test_diff_s3_restproxy(self):
        s = self.get_s3()
        r = RestProxy("local")
        #        
        test_diff(self, s, r)

    # Diff Cluster and X

    def test_diff_cluster_azureblob(self):
        c = Cluster("local")
        a = self.get_azureblob()
        #        
        test_diff(self, c, a)

    def test_diff_cluster_local(self):
        c = Cluster("local")
        l = self.get_local()
        #        
        test_diff(self, c, l)

    def test_diff_cluster_s3(self):
        c = Cluster("local")
        s = self.get_s3()
        #        
        test_diff(self, c, s)

    def test_diff_cluster_restproxy(self):
        c = Cluster("local")
        r = RestProxy("local")
        #        
        test_diff(self, c, r)

    # Diff RestProxy and X

    def test_diff_restproxy_azureblob(self):
        r = RestProxy("local")
        a = self.get_azureblob()
        #        
        test_diff(self, r, a)

    def test_diff_restproxy_local(self):
        r = RestProxy("local")
        l = self.get_local()
        #        
        test_diff(self, r, l)

    def test_diff_restproxy_s3(self):
        r = RestProxy("local")
        s = self.get_s3()
        #        
        test_diff(self, r, s)

    def test_diff_restproxy_cluster(self):
        r = RestProxy("local")
        c = Cluster("local")
        #        
        test_diff(self, r, c)

    # Files AzureBlob and X

    def test_from_to_file_azureblob_azureblob(self):
        a1 = self.get_azureblob()
        a2 = self.get_azureblob()
        #        
        test_from_to_file(self, a1, a2)

    def test_from_to_file_azureblob_local(self):
        a = self.get_azureblob()
        l = self.get_local()
        #        
        test_from_to_file(self, a, l)

    def test_from_to_file_azureblob_s3(self):
        a = self.get_azureblob()
        s = self.get_s3()
        #        
        test_from_to_file(self, a, s)

    # Files Local and X

    def test_from_to_file_local_azureblob(self):
        l = self.get_local()
        a = self.get_azureblob()
        #        
        test_from_to_file(self, l, a)

    def test_from_to_file_local_local(self):
        l1 = self.get_local()
        l2 = self.get_local()
        #        
        test_from_to_file(self, l1, l2)

    def test_from_to_file_local_s3(self):
        l = self.get_local()
        s = self.get_s3()
        #        
        test_from_to_file(self, l, s)

    # Files S3 and X

    def test_from_to_file_s3_azureblob(self):
        s = self.get_s3()
        a = self.get_azureblob()
        #        
        test_from_to_file(self, s, a)

    def test_from_to_file_s3_local(self):
        s = self.get_s3()
        l = self.get_local()
        #        
        test_from_to_file(self, s, l)

    def test_from_to_file_s3_s3(self):
        s1 = self.get_s3()
        s2 = self.get_s3()
        #        
        test_from_to_file(self, s1, s2)

    # Files Cluster and X

    def test_from_to_cluster_azureblob(self):
        c = Cluster("local")
        a = self.get_azureblob()
        #        
        test_from_to_file(self, c, a)

    def test_from_to_file_cluster_local(self):
        c = Cluster("local")
        l = self.get_local()
        #        
        test_from_to_file(self, c, l)

    def test_from_to_file_cluster_s3(self):
        c = Cluster("local")
        s = self.get_s3()
        #        
        test_from_to_file(self, c, s)

    # Files RestProxy and X

    def test_from_to_restproxy_azureblob(self):
        r = RestProxy("local")
        a = self.get_azureblob()
        #        
        test_from_to_file(self, r, a)

    def test_from_to_file_restproxy_local(self):
        r = RestProxy("local")
        l = self.get_local()
        #        
        test_from_to_file(self, r, l)

    def test_from_to_file_restproxy_s3(self):
        r = RestProxy("local")
        s = self.get_s3()
        #        
        test_from_to_file(self, r, s)

#

def test_cp(test_obj, storage1, storage2):
    partitions_int = 3
    # Create topic1 on storage1
    topic_str1 = test_obj.create_test_topic_name(storage1)
    storage1.create(topic_str1, partitions=partitions_int)
    random_int = random.randint(0, 2)
    type_str = ["avro", "protobuf", "jsonschema"][random_int]
    key_schema_str = test_obj.key_schema_str_list[random_int]
    value_schema_str = test_obj.value_schema_str_list[random_int]
    w = storage1.producer(topic_str1, key_type=type_str, value_type=type_str, key_schema=key_schema_str, value_schema=value_schema_str)
    #
    snack_dict_list = []
    for snack_dict in test_obj.snack_dict_list:
        snack_dict1 = snack_dict.copy()
        snack_dict1["name"] += "1"
        snack_dict2 = snack_dict.copy()
        snack_dict2["name"] += "2"
        snack_dict3 = snack_dict.copy()
        snack_dict3["name"] += "3"
        snack_dict_list += [snack_dict1, snack_dict2, snack_dict3]
    w.produce(snack_dict_list, key=[{"key": "1"}, {"key": "1"}, {"key": "1"}, {"key": "2"}, {"key": "2"}, {"key": "2"}, {"key": "3"}, {"key": "3"}, {"key": "3"}], headers=test_obj.headers_str_bytes_tuple_list)
    w.close()
    # Carbon copy topic1 on storage1 to topic2 on storage2 as bytes.
    topic_str2 = test_obj.create_test_topic_name(storage2)
    storage2.create(topic_str2, partitions=partitions_int)
    #
    group_str1 = test_obj.create_test_group_name(storage1)
    produce_batch_size_int1 = random.randint(1, 3*3)
    if storage1.__class__.__name__ == "RestProxy":
        # Need to use the native Kafka API here since the RestProxy V2 consumer does not support timestamps.
        c = Cluster("local")
        (consume_n_int1, written_n_int1) = c.cp(topic_str1, storage2, topic_str2, group=group_str1, source_key_type="bytes", source_value_type="bytes", target_key_type="bytes", target_value_type="bytes", produce_batch_size=produce_batch_size_int1, n=3*3, keep_timestamps=True, keep_partitions=True)
    else:
        (consume_n_int1, written_n_int1) = storage1.cp(topic_str1, storage2, topic_str2, group=group_str1, source_key_type="bytes", source_value_type="bytes", target_key_type="bytes", target_value_type="bytes", produce_batch_size=produce_batch_size_int1, n=3*3, keep_timestamps=True, keep_partitions=True)
    #
    test_obj.assertEqual(3*3, consume_n_int1)
    test_obj.assertEqual(3*3, written_n_int1)
    #
    # Carbon copy topic2 on storage2 back to topic3 on storage1 as bytes.
    topic_str3 = test_obj.create_test_topic_name(storage1)
    storage1.create(topic_str3, partitions=partitions_int)
    #
    group_str2 = test_obj.create_test_group_name(storage2)
    produce_batch_size_int2 = random.randint(1, 3*3)
    if storage2.__class__.__name__ == "RestProxy":
        # Need to use the native Kafka API here since the RestProxy V2 consumer does not support timestamps.
        c = Cluster("local")
        (consume_n_int2, written_n_int2) = c.cp(topic_str2, storage1, topic_str3, group=group_str2, source_key_type="bytes", source_value_type="bytes", target_key_type="bytes", target_value_type="bytes", produce_batch_size=produce_batch_size_int2, n=3*3, keep_timestamps=True, keep_partitions=True)
    else:
        (consume_n_int2, written_n_int2) = storage2.cp(topic_str2, storage1, topic_str3, group=group_str2, source_key_type="bytes", source_value_type="bytes", target_key_type="bytes", target_value_type="bytes", produce_batch_size=produce_batch_size_int2, n=3*3, keep_timestamps=True, keep_partitions=True)
    #
    test_obj.assertEqual(3*3, consume_n_int2)
    test_obj.assertEqual(3*3, written_n_int2)
    #
    # Have the partitions been carbon copied properly?
    storage1_topic1_watermarks_dict = storage1.watermarks(topic_str1)
    storage2_topic2_watermarks_dict = storage2.watermarks(topic_str2)
    storage1_topic3_watermarks_dict = storage1.watermarks(topic_str3)
    test_obj.assertEqual(list(storage1_topic1_watermarks_dict.values()), list(storage2_topic2_watermarks_dict.values()))
    test_obj.assertEqual(list(storage2_topic2_watermarks_dict.values()), list(storage1_topic3_watermarks_dict.values()))
    #
    # Have the timestamps been carbon copied properly?
    group_str3 = test_obj.create_test_group_name(storage1)
    if storage1.__class__.__name__ == "RestProxy":
        # Need to use the native Kafka API here since the RestProxy V2 consumer does not support timestamps.
        c = Cluster("local")
        message_dict_list1 = c.cat(topic_str1, group=group_str3, type=type_str, n=3*3)
    else:
        message_dict_list1 = storage1.cat(topic_str1, group=group_str3, type=type_str, n=3*3)
    #
    test_obj.assertEqual(3*3, len(message_dict_list1))
    #
    group_str4 = test_obj.create_test_group_name(storage1)
    if storage1.__class__.__name__ == "RestProxy":
        # Need to use the native Kafka API here since the RestProxy V2 consumer does not support timestamps.
        c = Cluster("local")
        message_dict_list2 = c.cat(topic_str3, group=group_str4, type=type_str, n=3*3)
    else:
        message_dict_list2 = storage1.cat(topic_str3, group=group_str4, type=type_str, n=3*3)
    #
    test_obj.assertEqual(3*3, len(message_dict_list2))
    #
    storage1_topic1_timestamp_set = set([message_dict["timestamp"] for message_dict in message_dict_list1])
    storage1_topic3_timestamp_set = set([message_dict["timestamp"] for message_dict in message_dict_list2])
    test_obj.assertEqual(storage1_topic1_timestamp_set, storage1_topic3_timestamp_set)

    # Copy topic3 on storage1 to topic5 on storage2 as json and do a tiny mapping.
    topic_str5 = test_obj.create_test_topic_name(storage2)
    storage2.create(topic_str5, partitions=partitions_int)
    #
    def map_ish(message_dict):
        message_dict["value"]["colour"] += "ish"
        return message_dict
    #

    group_str5 = test_obj.create_test_group_name(storage1)
    produce_batch_size_int3 = random.randint(1, 3*3)
    (consume_n_int3, written_n_int3) = storage1.cp(topic_str3, storage2, topic_str5, group=group_str5, source_type=type_str, target_type="json", produce_batch_size=produce_batch_size_int3, map_function=map_ish, n=3*3)
    #
    test_obj.assertEqual(3*3, consume_n_int3)
    test_obj.assertEqual(3*3, written_n_int3)
    #
    group_str6 = test_obj.create_test_group_name(storage1)
    message_dict_list3 = storage2.cat(topic_str5, group=group_str6, type="json", n=3*3)
    test_obj.assertEqual(3*3, len(message_dict_list3))
    #
    # Has the mapping been done properly?
    for message_dict in message_dict_list3:
        test_obj.assertTrue(message_dict["value"]["colour"].endswith("ish"))
    #
    # Has the order of the snacks been kept intact after all that copying?
    if storage1.__class__.__name__ != "RestProxy" and storage2.__class__.__name__ != "RestProxy":
        for i in range(3):
            j0 = next(j for j, message_dict in enumerate(message_dict_list3) if message_dict["value"]["name"] == snack_dict_list[3*i]["name"])
            j1 = next(j for j, message_dict in enumerate(message_dict_list3) if message_dict["value"]["name"] == snack_dict_list[3*i+1]["name"])
            j2 = next(j for j, message_dict in enumerate(message_dict_list3) if message_dict["value"]["name"] == snack_dict_list[3*i+2]["name"])
            #
            test_obj.assertEqual(message_dict_list3[j0]["partition"], message_dict_list3[j1]["partition"])
            test_obj.assertEqual(message_dict_list3[j1]["partition"], message_dict_list3[j2]["partition"])
            #
            test_obj.assertLess(message_dict_list3[j0]["offset"], message_dict_list3[j1]["offset"])
            test_obj.assertLess(message_dict_list3[j1]["offset"], message_dict_list3[j2]["offset"])

#

def test_diff(test_obj, storage1, storage2):
    topic_str1 = test_obj.create_test_topic_name(storage1)
    storage1.create(topic_str1)
    w1 = storage1.producer(topic_str1, type="str")
    w1.produce(test_obj.snack_str_list)
    w1.close()
    #
    topic_str2 = test_obj.create_test_topic_name(storage2)
    storage2.create(topic_str2)
    w2 = storage2.producer(topic_str2, type="str")
    w2.produce(test_obj.snack_ish_dict_list)
    w2.close()
    #
    group_str1 = test_obj.create_test_group_name(storage1)
    time.sleep(0.1)
    group_str2 = test_obj.create_test_group_name(storage2)
    #
    (message_dict_message_dict_tuple_list, message_counter_int1, message_counter_int2) = storage1.diff(topic_str1, storage2, topic_str2, source1_group=group_str1, source2_group=group_str2, source1_type="json", source2_type="json", n=3, consume_batch_size=3)
    test_obj.assertEqual(3, len(message_dict_message_dict_tuple_list))
    test_obj.assertEqual(3, message_counter_int1)
    test_obj.assertEqual(3, message_counter_int2)

# Files

def test_from_to_file(test_obj, storage1, storage2):
    if storage2.__class__.__name__ in ["AzureBlob", "Local", "S3"]:
        suffix_str_list = [".csv", ".feather", ".json", ".orc", ".parquet", ".xlsx", ".xml"]
        random_int = random.randint(0, len(suffix_str_list) - 1)
        suffix_str = suffix_str_list[random_int]
        #
        topic_str1 = test_obj.create_test_topic_name(storage1)
        storage1.create(topic_str1)
        producer = storage1.producer(topic_str1, value_type="json")
        producer.produce(test_obj.snack_str_list)
        producer.close()
        #
        group_str1 = test_obj.create_test_group_name(storage1)
        file_str = f"{topic_str1}{suffix_str}"
        n_int1 = storage1.topic_to_file(topic_str1, storage2, file_str, group=group_str1, type="json")
        test_obj.assertTrue(n_int1, 3)
        #
        topic_str2 = test_obj.create_test_topic_name(storage1)
        storage1.create(topic_str2)
        storage2.file_to_topic(file_str, storage1, topic_str2)
        #
        group_str2 = test_obj.create_test_group_name(storage1)
        message_dict_list = storage1.cat(topic_str2, group=group_str2, n=3, type="json")
        test_obj.assertEqual(3, len(message_dict_list))
        test_obj.assertEqual(500.0, message_dict_list[0]["value"]["calories"])
        test_obj.assertEqual(260.0, message_dict_list[1]["value"]["calories"])
        test_obj.assertEqual(80.0, message_dict_list[2]["value"]["calories"])
        #
        storage2.admin.delete_file(storage2.admin.get_abs_path_str(f"files/{file_str}"))



from kafi.deserializer import Deserializer
from kafi.helpers import get_millis

class StorageConsumer(Deserializer):
    def __init__(self, storage_obj, *topics, **kwargs):
        self.storage_obj = storage_obj
        #
        super().__init__(storage_obj.schema_registry_config_dict)
        #
        # Get topics to subscribe to.
        self.topic_str_list = list(topics)
        #
        # Get start offsets for the subscribed topics.
        self.topic_str_start_offsets_dict_dict = self.get_start_offsets_from_kwargs(self.topic_str_list, **kwargs)
        #
        # Get key and value types.
        (self.topic_str_key_type_str_dict, self.topic_str_value_type_str_dict) = self.get_key_value_types_from_kwargs(self.topic_str_list, **kwargs)
        #
        # Get group.
        self.group_str = self.get_group_str_from_kwargs(**kwargs)
        #
        # Configure.
        self.consumer_config_dict = {}
        self.consumer_config_dict["auto.offset.reset"] = storage_obj.auto_offset_reset()
        if "config" in kwargs:
            for key_str, value in kwargs["config"].items():
                self.consumer_config_dict[key_str] = value
        #
        self.enable_auto_commit_bool = kwargs["enable_auto_commit"] if "enable_auto_commit" in kwargs else storage_obj.enable_auto_commit()
        #
        self.schema_id_int_generalizedProtocolMessageType_protobuf_schema_str_tuple_dict = {}

    #

    def get_start_offsets_from_kwargs(self, topic_str_list, **kwargs):
        if "offsets" in kwargs and kwargs["offsets"] is not None:
            offsets_dict = kwargs["offsets"]
            first_key_str_or_int = list(offsets_dict.keys())[0]
            if isinstance(first_key_str_or_int, int):
                topic_str_offsets_dict_dict = {topic_str: offsets_dict for topic_str in topic_str_list}
            else:
                topic_str_offsets_dict_dict = offsets_dict
        else:
            topic_str_offsets_dict_dict = None
        #
        if topic_str_offsets_dict_dict is not None:
            offset_int_list = sum({topic_str: list(offsets_dict.values()) for topic_str, offsets_dict in topic_str_offsets_dict_dict.items()}.values(), [])
            if any(offset_int < 0 for offset_int in offset_int_list):
                topic_str_partition_int_size_int_dict_dict = self.storage_obj.ls(topic_str_list, partitions=True)
                topic_str_offsets_dict_dict = {topic_str: {partition_int: (topic_str_partition_int_size_int_dict_dict[topic_str][partition_int] + offset_int if offset_int < 0 else offset_int) for partition_int, offset_int in offsets_dict.items()} for topic_str, offsets_dict in topic_str_offsets_dict_dict.items()}
        #
        return topic_str_offsets_dict_dict

    def get_key_value_types_from_kwargs(self, topic_str_list, **kwargs):
        (key_type_str, value_type_str) = self.storage_obj.get_key_value_type_tuple(**kwargs)
        #
        topic_str_key_type_str_dict = {topic_str: key_type_str for topic_str in topic_str_list}
        topic_str_value_type_str_dict = {topic_str: value_type_str for topic_str in topic_str_list}
        #
        return (topic_str_key_type_str_dict, topic_str_value_type_str_dict)

    #

    def get_group_str_from_kwargs(self, **kwargs):
        if "group" in kwargs:
            return kwargs["group"]
        else:
            prefix_str = self.storage_obj.consumer_group_prefix()
            return prefix_str + str(get_millis())



import io
import os

from kafi.fs.fs_admin import FSAdmin

#

class S3Admin(FSAdmin):
    def __init__(self, s3_obj):
        from minio import Minio
        #

        super().__init__(s3_obj)
        #
        self.minio = Minio(s3_obj.s3_config_dict["endpoint"], access_key=s3_obj.s3_config_dict["access.key"], secret_key=s3_obj.s3_config_dict["secret.key"], secure=False)

    # Topics/Files

    def list_dirs(self, abs_path_dir_str):
        object_generator = self.minio.list_objects(self.storage_obj.bucket_name(), prefix=abs_path_dir_str, recursive=True)
        rel_dir_str_set = set([os.path.basename(os.path.dirname(object.object_name)) for object in object_generator])
        #
        rel_dir_str_list = list(rel_dir_str_set)
        rel_dir_str_list.sort()
        #
        return rel_dir_str_list

    def list_files(self, abs_path_dir_str):
        object_generator = self.minio.list_objects(self.storage_obj.bucket_name(), prefix=abs_path_dir_str, recursive=True)
        rel_file_str_set = set([os.path.relpath(object.object_name, abs_path_dir_str) for object in object_generator])
        #
        rel_file_str_list = list(rel_file_str_set)
        rel_file_str_list.sort()
        #
        return rel_file_str_list

    def delete_file(self, abs_path_file_str):
        self.minio.remove_object(self.storage_obj.bucket_name(), abs_path_file_str)

    def delete_dir(self, _):
        pass

    def exists_file(self, abs_path_file_str):
        from minio.error import MinioException
        #

        try:
            self.minio.stat_object(self.storage_obj.bucket_name(), abs_path_file_str)
            return True
        except MinioException:
            return False

    # Metadata
    
    def read_str(self, abs_path_file_str):
        response = self.minio.get_object(self.storage_obj.bucket_name(), abs_path_file_str)
        object_bytes = response.data
        #
        object_str = object_bytes.decode("utf-8")
        #
        return object_str

    def write_str(self, abs_path_file_str, data_str):
        data_bytes = data_str.encode("utf-8")
        #
        self.minio.put_object(self.storage_obj.bucket_name(), abs_path_file_str, io.BytesIO(data_bytes), length=len(data_bytes))

    #

    def read_bytes(self, abs_path_file_str):
        from minio import Minio
        #

        self.minio = Minio(self.storage_obj.s3_config_dict["endpoint"], access_key=self.storage_obj.s3_config_dict["access.key"], secret_key=self.storage_obj.s3_config_dict["secret.key"], secure=False)
        #
        response = self.minio.get_object(self.storage_obj.bucket_name(), abs_path_file_str)
        object_bytes = response.data
        #
        return object_bytes

    def write_bytes(self, abs_path_file_str, data_bytes):
        from minio import Minio
        #

        self.minio = Minio(self.storage_obj.s3_config_dict["endpoint"], access_key=self.storage_obj.s3_config_dict["access.key"], secret_key=self.storage_obj.s3_config_dict["secret.key"], secure=False)
        #
        self.minio.put_object(self.storage_obj.bucket_name(), abs_path_file_str, io.BytesIO(data_bytes), length=len(data_bytes))



from kafi.helpers import pattern_match

#

class StorageAdmin():
    def __init__(self, storage_obj, **kwargs):
        self.storage_obj = storage_obj

    #

    def topics(self, pattern=None, size=False, **kwargs):
        pattern_str_or_str_list = pattern
        size_bool = size
        partitions_bool = "partitions" in kwargs and kwargs["partitions"]
        #
        timeout_int = kwargs["timeout"] if "timeout" in kwargs else -1.0
        #
        def size(pattern_str_or_str_list, timeout_int):
            topic_str_partition_int_tuple_dict_dict = self.watermarks(pattern_str_or_str_list, timeout=timeout_int)
            #
            topic_str_size_int_partitions_dict_tuple_dict = {}
            for topic_str, partition_int_tuple_dict in topic_str_partition_int_tuple_dict_dict.items():
                partitions_dict = {partition_int: partition_int_tuple_dict[partition_int][1]-partition_int_tuple_dict[partition_int][0] for partition_int in partition_int_tuple_dict.keys()}
                #
                size_int = 0
                for offset_int_tuple in partition_int_tuple_dict.values():
                    partition_size_int = offset_int_tuple[1] - offset_int_tuple[0]
                    size_int += partition_size_int
                #
                topic_str_size_int_partitions_dict_tuple_dict[topic_str] = (size_int, partitions_dict)
            return topic_str_size_int_partitions_dict_tuple_dict
        #

        if size_bool:
            topic_str_size_int_partitions_dict_tuple_dict = size(pattern_str_or_str_list, timeout_int)
            if partitions_bool:
                # e.g. {"topic": {"size": 42, "partitions": {0: 23, 1: 4711}}}
                topic_str_size_int_partitions_dict_dict = {topic_str: {"size": size_int_partitions_dict_tuple[0], "partitions": size_int_partitions_dict_tuple[1]} for topic_str, size_int_partitions_dict_tuple in topic_str_size_int_partitions_dict_tuple_dict.items()}
                return topic_str_size_int_partitions_dict_dict
            else:
                # e.g. {"topic": 42}
                topic_str_size_int_dict = {topic_str: size_int_partitions_dict_tuple[0] for topic_str, size_int_partitions_dict_tuple in topic_str_size_int_partitions_dict_tuple_dict.items()}
                return topic_str_size_int_dict
        else:
            if partitions_bool:
                # e.g. {"topic": {0: 23, 1: 4711}}
                topic_str_size_int_partitions_dict_tuple_dict = size(pattern_str_or_str_list, timeout_int)
                topic_str_partitions_dict_dict = {topic_str: size_int_partitions_dict_tuple[1] for topic_str, size_int_partitions_dict_tuple in topic_str_size_int_partitions_dict_tuple_dict.items()}
                return topic_str_partitions_dict_dict
            else:
                # e.g. ["topic"]
                topic_str_list = self.list_topics(pattern_str_or_str_list)
                filtered_topic_str_list = pattern_match(topic_str_list, pattern_str_or_str_list)
                return filtered_topic_str_list



import os
import sys

if os.path.basename(os.getcwd()) == "test":
    sys.path.insert(1, "..")
else:
    sys.path.insert(1, ".")

from test.test_single_storage_base import TestSingleStorageBase
from kafi.kafka.cluster.cluster import Cluster

#

class Test(TestSingleStorageBase):
    def setUp(self):
        super().setUp()
        self.principal_str = None
        # self.principal_str = "User:admin"


    def tearDown(self):
        super().tearDown()

#

    def get_storage(self):
        c = Cluster("local")
#        c = Cluster("ccloud")
        #
        return c

    def is_ccloud(self):
        c = self.get_storage()
        return "confluent.cloud" in c.kafka_config_dict["bootstrap.servers"]



import base64
import binascii
import datetime
import dateutil.parser
from fnmatch import fnmatch
from functools import reduce
import json
# import logging
import pandas as pd
import requests
from requests.adapters import HTTPAdapter, Retry
import sys
import time

def get_millis():
    return int(time.time()*1000)


def to_millis(timestamp_str):
    return int(dateutil.parser.isoparse(timestamp_str).timestamp()*1000)


def from_millis(millis_int):
    return datetime.datetime.fromtimestamp(millis_int/1000.0).isoformat(sep=" ")


def is_interactive():
    return hasattr(sys, 'ps1')


def pretty(dict):
    return json.dumps(dict, indent=2, default=str)


def ppretty(dict):
    print(pretty(dict))

#

def create_session(retries_int):
    # logging.basicConfig(level=logging.DEBUG)
    session = requests.Session()
    retry = Retry(total=retries_int, backoff_factor=1, status_forcelist=[500, 502, 503, 504], allowed_methods=None)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    #
    return session


def get(url_str, headers_dict, payload_dict=None, auth_str_tuple=None, retries_int=0, debug_bool=False):
    session = create_session(retries_int)
    #
    if payload_dict is None:
        if debug_bool:
            print(f"GET Request\n-\nURL: {url_str}\nHeaders: {headers_dict}\n")
        #
        response = session.get(url_str, headers=headers_dict, auth=auth_str_tuple)
    else:
        if debug_bool:
            print(f"GET Request\n-\nURL: {url_str}\nHeaders: {headers_dict}\nPayload: {payload_dict}")
        #
        response = session.get(url_str, headers=headers_dict, json=payload_dict, auth=auth_str_tuple)
    #
    if debug_bool:
        print(f"GET Response\n-\n{response.text}\n")
    #
    if is_json(response.text):
        response_dict = response.json()
    else:
        response_dict = {"response": response.text}
    #
    if isinstance(response_dict, dict):
        if "error_code" in response_dict and response_dict["error_code"] > 400:
            raise Exception(response_dict["message"])
    #
    if response.ok:
        return response_dict
    else:
        raise Exception(response_dict)


def delete(url_str, headers_dict, auth_str_tuple=None, retries_int=10, debug_bool=False):
    session = create_session(retries_int)
    #
    if debug_bool:
        print(f"DELETE Request\n-\nURL: {url_str}\nHeaders: {headers_dict}\n")
    #
    response = session.delete(url_str, headers=headers_dict, auth=auth_str_tuple)
    #
    if debug_bool:
        print(f"DELETE Response\n-\n{response.text}\n")
    #
    if is_json(response.text):
        response_dict = response.json()
    else:
        response_dict = {"response": response.text}
    #
    if isinstance(response_dict, dict):
        if "error_code" in response_dict and response_dict["error_code"] > 400:
            raise Exception(response_dict["message"])
    #
    if response.ok:
        return response_dict
    else:
        raise Exception(response_dict)


def post(url_str, headers_dict, payload_dict_or_generator, auth_str_tuple=None, retries_int=10, debug_bool=False):
    session = create_session(retries_int)
    #
    if isinstance(payload_dict_or_generator, dict):
        if debug_bool:
            print(f"POST Request\n-\nURL: {url_str}\nHeaders: {headers_dict}\nPayload: {payload_dict_or_generator}\n")
        #
        response = session.post(url_str, headers=headers_dict, json=payload_dict_or_generator, auth=auth_str_tuple)
    else:
        if debug_bool:
            print(f"POST Request\n-\nURL: {url_str}\nHeaders: {headers_dict}\nPayload: (generator)\n")
        #
        response = session.post(url_str, headers=headers_dict, data=payload_dict_or_generator, auth=auth_str_tuple)
    #
    if debug_bool:
        print(f"POST Response\n-\n{response.text}\n")
    #
    if is_json(response.text):
        response_dict = response.json()
        #
        if isinstance(response_dict, dict):
            if "error_code" in response_dict and response_dict["error_code"] > 400:
                raise Exception(response_dict["message"])
            #
        if response.ok:
            return response_dict
        else:
            raise Exception(response_dict)
    else:
        response_text_list = "[" + response.text[:-2].replace("\r\n", ",") + "]"
        response_dict_list = json.loads(response_text_list)
        #
        for response_dict in response_dict_list:
            if isinstance(response_dict, dict):
                if "error_code" in response_dict and response_dict["error_code"] > 400:
                    raise Exception(response_dict["message"])
        #
        if response.ok:
            return response_dict_list
        else:
            raise Exception(response_dict_list)


def get_auth_str_tuple(basic_auth_user_info):
    if basic_auth_user_info is None:
        auth_str_tuple = None
    else:
        auth_str_tuple = tuple(basic_auth_user_info.split(":"))
    #
    return auth_str_tuple


def is_json(str):
    try:
        json.loads(str)
    except ValueError as e:
        return False
    return True


def is_pattern(str):
    return "*" in str or "?" in str or ("[" in str and "]" in str) or ("[!" in str and "]" in str)


def is_base64_encoded(str_or_bytes_or_dict):
    try:
        if isinstance(str_or_bytes_or_dict, bytes):
            decoded_bytes = base64.b64decode(str_or_bytes_or_dict)
        elif isinstance(str_or_bytes_or_dict, str):
            decoded_bytes = base64.b64decode(bytes(str_or_bytes_or_dict, encoding="utf-8"))
        elif isinstance(str_or_bytes_or_dict, dict):
            decoded_bytes = base64.b64decode(bytes(json.dumps(str_or_bytes_or_dict, default=str), encoding="utf-8"))
        else:
            return False
        encoded_bytes = base64.b64encode(decoded_bytes)
        return str_or_bytes_or_dict == encoded_bytes
    except (binascii.Error, UnicodeDecodeError):
        return False


def base64_encode(str_or_bytes_or_dict):
    if isinstance(str_or_bytes_or_dict, bytes):
        encoded_bytes = base64.b64encode(str_or_bytes_or_dict)
    elif isinstance(str_or_bytes_or_dict, str):
        encoded_bytes = base64.b64encode(bytes(str_or_bytes_or_dict, encoding="utf-8"))
    elif isinstance(str_or_bytes_or_dict, dict):
        encoded_bytes = base64.b64encode(bytes(json.dumps(str_or_bytes_or_dict, default=str), encoding="utf-8"))
    return encoded_bytes


def base64_decode(base64_str):
    return base64.b64decode(bytes(base64_str, encoding="utf-8"))


def to_bytes(data):
    if isinstance(data, bytes):
        data_bytes = data
    elif isinstance(data, str):
        data_bytes = data.encode("utf-8")
    elif isinstance(data, dict):
        data_bytes = json.dumps(data, default=str).encode("utf-8")
    elif data is None:
        data_bytes = None
    else:
        data_str = str(data)
        data_bytes = data_str.encode("utf-8")
    #
    return data_bytes


def bytes_or_str_to_bytes(bytes_or_str):
    if isinstance(bytes_or_str, bytes):
        return_bytes = bytes_or_str
    elif isinstance(bytes_or_str, str):
        return_bytes = bytes(bytes_or_str, encoding="utf-8")
    #
    return return_bytes


def bytes_to_str(bytes):
    if bytes is None:
        str = None
    else:
        str = bytes.decode("utf-8")
    #
    return str


def bytes_to_dict(bytes):
    if bytes is None:
        dict = None
    else:
        dict = json.loads(bytes.decode("utf-8"))
    #
    return dict


def str_to_bytes(str):
    if str is None:
        bytes = None
    else:
        bytes = str.encode("utf-8")
    #
    return bytes


def pattern_match(input_str_list, pattern_str_or_str_list):
    if pattern_str_or_str_list is not None:
        if isinstance(pattern_str_or_str_list, str):
            pattern_str_or_str_list = [pattern_str_or_str_list]
        output_topic_str_list = [input_str for input_str in input_str_list if any(fnmatch(input_str, pattern_str) for pattern_str in pattern_str_or_str_list)]
    else:
        output_topic_str_list = input_str_list
    #
    output_topic_str_list.sort()
    #
    return output_topic_str_list


def explode_normalize(df):
    def explode(df, col_str):
        df = df.explode(col_str)
        #
        if isinstance(df.iloc[0][col_str], list):
            df = explode(df, col_str)
        elif isinstance(df.iloc[0][col_str], object):
            df_child = pd.json_normalize(df[col_str])
            df_child.columns = [f'{col_str}.{child_col_str}' for child_col_str in df_child.columns]
            df = pd.concat([df.loc[:, ~df.columns.isin([col_str])].reset_index(drop=True), df_child], axis=1)
        #
        return df
    #
    col_str_list = [col_str for col_str in df.columns if isinstance(df.iloc[0][col_str], list)]
    if len(col_str_list) < 1:
        return df
    #
    df = reduce(explode, col_str_list, df)
    #
    return df


def zip2(list1, list2):
    i = 0
    j = 0
    zipped_list = []
 
    while True:
        if len(list1) >= len(list2):
            if i < len(list1):
                zipped_list.append((list1[i], list2[j]))
                i += 1
                j += 1
            else:
                break
            if j >= len(list2):
                j = 0
        else:
            if j < len(list2):
                zipped_list.append((list1[i], list2[j]))
                i += 1
                j += 1
            else:
                break
            if i >= len(list1):
                i = 0
    
    return zipped_list



from kafi.fs.fs_consumer import FSConsumer

# Constants

ALL_MESSAGES = -1

#

class LocalConsumer(FSConsumer):
    def __init__(self, local_obj, topic, **kwargs):
        super().__init__(local_obj, topic, **kwargs)
    
    #

    def close(self):
        return self.topic_str_list



import ast
from fnmatch import fnmatch
import os

from kafi.storage_admin import StorageAdmin
from kafi.helpers import get_millis, pattern_match

class FSAdmin(StorageAdmin):
    def __init__(self, fs_obj, **kwargs):
        super().__init__(fs_obj, **kwargs)
        #
        self.default_state_str = "stable"

    #

    def list_topics(self, pattern=None):
        root_dir_str = self.storage_obj.root_dir()
        rel_dir_str_list = self.list_dirs(os.path.join(root_dir_str, "topics"))
        #
        all_topic_str_list = [rel_dir_str for rel_dir_str in rel_dir_str_list if not rel_dir_str.endswith("partitions")]
        #
        topic_or_file_str_list = pattern_match(all_topic_str_list, pattern)
        #
        return topic_or_file_str_list

    def list_groups(self, pattern=None):
        root_dir_str = self.storage_obj.root_dir()
        rel_file_str_list = self.list_files(os.path.join(root_dir_str, "groups"))
        #
        all_group_str_list = [rel_file_str for rel_file_str in rel_file_str_list]
        all_group_str_set = set(all_group_str_list)
        all_group_str_list = list(all_group_str_set)
        #
        group_str_list = pattern_match(all_group_str_list, pattern)
        #
        return group_str_list

    #

    def config(self, pattern, config=None, **kwargs):
        config_dict = config
        #
        topic_str_list = self.list_topics(pattern)
        #
        if config_dict is not None:
            for topic_str in topic_str_list:
                metadata_dict = self.get_metadata(topic_str)
                metadata_dict["config"] = config_dict
                self.set_metadata(topic_str, metadata_dict)
        #
        topic_str_config_dict_dict = {topic_str: self.get_config(topic_str) for topic_str in topic_str_list}
        #
        return topic_str_config_dict_dict

    #

    def create(self, topic, partitions=1, config={}, **kwargs):
        topic_str = topic
        #
        if self.list_topics(topic_str) != []:
            raise Exception(f"Topic \"{topic_str}\" already exists.")
        #
        config_dict = config
        partitions_int = partitions
        #
        metadata_dict = {"topic": topic_str, "partitions": partitions_int, "config": config_dict}
        self.set_metadata(topic_str, metadata_dict)
    
    #

    def delete(self, pattern, **kwargs):
        topic_str_list = self.list_topics(pattern)
        #
        for topic_str in topic_str_list:
            topic_abs_dir_str = self.get_topic_abs_path_str(topic_str)
            #
            rel_file_str_list = self.list_files(topic_abs_dir_str)
            for rel_file_str in rel_file_str_list:
                self.delete_file(os.path.join(topic_abs_dir_str, rel_file_str))
            #
            rel_dir_str_list = self.list_dirs(topic_abs_dir_str)
            for rel_dir_str in rel_dir_str_list:
                self.delete_dir(os.path.join(topic_abs_dir_str, rel_dir_str))
            #
            self.delete_dir(topic_abs_dir_str)
        #
        return topic_str_list
    
    #

    def offsets_for_times(self, pattern, partitions_timestamps, **kwargs):
        pattern_str_or_str_list = pattern
        partition_int_timestamp_int_dict = partitions_timestamps
        #
        topic_str_list = self.list_topics(pattern_str_or_str_list)
        #
        topic_str_partition_int_offsets_int_dict_dict = {}
        for topic_str in topic_str_list:
            partitions_int = self.get_partitions(topic_str)
            #
            abs_topic_dir_str = self.storage_obj.admin.get_topic_abs_path_str(topic_str)
            #
            for partition_int in range(partitions_int):
                rel_file_str = self.storage_obj.admin.find_partition_file_str_by_timestamp(topic_str, partition_int, partition_int_timestamp_int_dict[partition_int])
                #
                messages_bytes = self.storage_obj.admin.read_bytes(os.path.join(abs_topic_dir_str, "partitions", rel_file_str))
                #
                message_bytes_list = messages_bytes.split(b"\n")[:-1]
                #
                for message_bytes in message_bytes_list:
                    message_dict = ast.literal_eval(message_bytes.decode("utf-8"))
                    #
                    if message_dict["timestamp"][1] >= partition_int_timestamp_int_dict[partition_int]:
                        if topic_str not in topic_str_partition_int_offsets_int_dict_dict:
                            topic_str_partition_int_offsets_int_dict_dict[topic_str] = {}
                        #
                        topic_str_partition_int_offsets_int_dict_dict[topic_str][partition_int] = message_dict["offset"]
                        break
        #
        return topic_str_partition_int_offsets_int_dict_dict

    #

    def partitions(self, pattern=None, partitions=None, verbose=False, **kwargs):
        partitions_int = partitions
        #
        topic_str_list = self.list_topics(pattern)
        #
        if partitions_int is not None:
            topic_str_partitions_int_dict = {}
            for topic_str in topic_str_list:
                metadata_dict = self.get_metadata(topic_str)
                #
                metadata_dict["partitions"] = partitions_int
                #
                self.set_metadata(topic_str, metadata_dict)
                #
                topic_str_partitions_int_dict[topic_str] = partitions_int
        #
        topic_str_partitions_int_dict = {topic_str: self.get_partitions(topic_str) for topic_str in topic_str_list}
        #
        return topic_str_partitions_int_dict

    #

    def watermarks(self, pattern, **kwargs):
        topic_str_list = self.list_topics(pattern)
        filtered_topic_str_list = pattern_match(topic_str_list, pattern)
        #
        def get_watermark_offsets(topic_str, partition_int):
            topic_abs_dir_str = self.get_topic_abs_path_str(topic_str)
            rel_file_str_list = self.list_files(os.path.join(topic_abs_dir_str, "partitions"))
            partition_rel_file_str_list = [rel_file_str for rel_file_str in rel_file_str_list if int(rel_file_str.split(",")[0]) == partition_int]
            partition_rel_file_str_list.sort()
            low_offset_int = 0
            high_offset_int = 0
            if len(partition_rel_file_str_list) > 0:
                first_partition_rel_file_str = partition_rel_file_str_list[0]
                last_partition_rel_file_str = partition_rel_file_str_list[-1]
                #
                low_offset_int = int(first_partition_rel_file_str.split(",")[1])
                high_offset_int = int(last_partition_rel_file_str.split(",")[2]) + 1
            #
            return (low_offset_int, high_offset_int)
        #

        topic_str_partition_int_offsets_tuple_dict_dict = {}
        for topic_str in filtered_topic_str_list:
            partitions_int = self.get_partitions(topic_str)
            partition_int_offsets_tuple_dict = {partition_int: get_watermark_offsets(topic_str, partition_int) for partition_int in range(partitions_int)}
            topic_str_partition_int_offsets_tuple_dict_dict[topic_str] = partition_int_offsets_tuple_dict
        #
        return topic_str_partition_int_offsets_tuple_dict_dict

    #
    # File-based topics
    #

    # Topics/Partitions

    def get_abs_path_str(self, rel_path_str):
        abs_path_str = os.path.join(self.storage_obj.root_dir(), rel_path_str)
        #
        return abs_path_str

    def get_topic_abs_path_str(self, topic_str):
        topic_abs_dir_str = os.path.join(self.storage_obj.root_dir(), "topics", topic_str)
        #
        return topic_abs_dir_str

    def get_file_abs_path_str(self, file_str):
        file_abs_file_str = os.path.join(self.storage_obj.root_dir(), "files", file_str)
        #
        return file_abs_file_str

    def find_partition_file_str_by_offset(self, topic_str, partition_int, to_find_offset_int):
        # Get sorted list of all relative file names rel_file_str_list for the partition files for partition_int of topic_str.
        topic_abs_dir_str = self.get_topic_abs_path_str(topic_str)
        rel_file_str_list1 = self.list_files(os.path.join(topic_abs_dir_str, "partitions"))
        rel_file_str_list = [rel_file_str for rel_file_str in rel_file_str_list1 if int(rel_file_str.split(",")[0]) == partition_int]
        if rel_file_str_list == []:
            return None
        rel_file_str_list.sort()
        # Now find the first file in which the message with offset to_find_offset_int is contained :)
        rel_file_str_start_offset_int_end_offset_int_tuple_dict = {rel_file_str: (int(rel_file_str.split(",")[1]), int(rel_file_str.split(",")[2])) for rel_file_str in rel_file_str_list}
        found_rel_file_str = None
        for rel_file_str, start_offset_int_end_offset_int_tuple in rel_file_str_start_offset_int_end_offset_int_tuple_dict.items():
            (start_offset_int, end_offset_int) = start_offset_int_end_offset_int_tuple
            #
            if to_find_offset_int >= start_offset_int and to_find_offset_int <= end_offset_int:
                found_rel_file_str = rel_file_str
                break
        #
        return found_rel_file_str

    def find_partition_file_str_by_timestamp(self, topic_str, partition_int, to_find_timestamp_int):
        # Get sorted list of all relative file names rel_file_str_list for the partition files for partition_int of topic_str.
        topic_abs_dir_str = self.get_topic_abs_path_str(topic_str)
        rel_file_str_list1 = self.list_files(os.path.join(topic_abs_dir_str, "partitions"))
        rel_file_str_list = [rel_file_str for rel_file_str in rel_file_str_list1 if int(rel_file_str.split(",")[0]) == partition_int]
        if rel_file_str_list == []:
            return None
        rel_file_str_list.sort()
        # Now find the first file in which the message with timestamp to_find_timestamp_int is be contained :)
        rel_file_str_start_timestamp_int_end_timestamp_int_tuple_dict = {rel_file_str: (int(rel_file_str.split(",")[3]), int(rel_file_str.split(",")[4])) for rel_file_str in rel_file_str_list}
        found_rel_file_str = None
        for rel_file_str, start_timestamp_int_end_timestamp_int_tuple in rel_file_str_start_timestamp_int_end_timestamp_int_tuple_dict.items():
            (start_timestamp_int, end_timestamp_int) = start_timestamp_int_end_timestamp_int_tuple
            #
            if to_find_timestamp_int >= start_timestamp_int and to_find_timestamp_int <= end_timestamp_int:
                found_rel_file_str = rel_file_str
                break
        #
        return found_rel_file_str

    def get_partition_files(self, topic_str):
        topic_abs_dir_str = self.get_topic_abs_path_str(topic_str)
        #
        partitions_int = self.get_partitions(topic_str)
        #
        rel_file_str_list = self.list_files(os.path.join(topic_abs_dir_str, "partitions"))
        #
        def sort(list):
            list.sort()
            return list
        #

        partition_int_rel_file_str_list_dict = {partition_int: sort([rel_file_str for rel_file_str in rel_file_str_list if int(rel_file_str.split(",")[0]) == partition_int]) for partition_int in range(partitions_int)}
        #
        return partition_int_rel_file_str_list_dict

    #

    def delete_groups(self, pattern, state_pattern="*"):
        group_str_list = self.groups(pattern, state_pattern)
        #
        root_dir_str = self.storage_obj.root_dir()
        for group_str in group_str_list:
            group_abs_path_file_str = os.path.join(root_dir_str, "groups", f"{group_str}")
            #
            self.delete_file(group_abs_path_file_str)
        #
        return group_str_list

    def describe_groups(self, pattern="*", state_pattern="*"):
        group_str_state_str_dict = self.groups(pattern, state_pattern, state=True)
        #
        group_str_group_description_dict_dict = {group_str: {"group_id": group_str, "is_simple_consumer_group": False, "partition_assignor": "range", "state": state_str} for group_str, state_str in group_str_state_str_dict.items()}
        #

        return group_str_group_description_dict_dict

    def groups(self, pattern="*", state_pattern="*", state=False):
        state_pattern_str_list = [state_pattern] if isinstance(state_pattern, str) else state_pattern
        state_bool = state
        #
        group_str_list = self.list_groups(pattern)
        #
        group_str_state_str_tuple_list = [(group_str, self.get_group_state(group_str)) for group_str in group_str_list]
        group_str_state_str_tuple_list = [(group_str, state_str) for group_str, state_str in group_str_state_str_tuple_list if any(fnmatch(state_str, state_pattern_str) for state_pattern_str in state_pattern_str_list)]
        #
        if state_bool:
            group_str_state_str_dict = {group_str: state_str for group_str, state_str in group_str_state_str_tuple_list}
            return group_str_state_str_dict
        else:
            group_str_list = [group_str for group_str, _ in group_str_state_str_tuple_list]
            return group_str_list

    def group_offsets(self, pattern, group_offsets=None, state_pattern="*"):
        topic_str_offsets_dict_dict = group_offsets
        #
        group_str_list = self.groups(pattern, state_pattern)
        #
        if topic_str_offsets_dict_dict is not None:
            for group_str in group_str_list:
                new_group_dict = {"offsets": topic_str_offsets_dict_dict}
                self.set_group_dict(group_str, new_group_dict)
        #
        group_str_topic_str_offsets_dict_dict_dict = {group_str: self.get_group_dict(group_str)["offsets"] for group_str in group_str_list}
        #
        return group_str_topic_str_offsets_dict_dict_dict

    # Metadata/Groups

    def read_dict_from_file(self, abs_path_file_str):
        if self.exists_file(abs_path_file_str):
            data_str = self.read_str(abs_path_file_str)
            #
            if data_str is not None:
                data_dict = ast.literal_eval(data_str)
            else:
                data_dict = {}
        else:
            data_dict = {}
        #
        return data_dict

    def write_dict_to_file(self, abs_path_file_str, data_dict):
        data_str = str(data_dict)
        #
        self.write_str(abs_path_file_str, data_str)

    # Metadata

    def get_metadata(self, topic_str):
        topic_dir_str = self.get_topic_abs_path_str(topic_str)
        metadata_dict = self.read_dict_from_file(os.path.join(topic_dir_str, "metadata"))
        #
        return metadata_dict

    def get_partitions(self, topic_str):
        metadata_dict = self.get_metadata(topic_str)
        partitions_int = metadata_dict["partitions"]
        #
        return partitions_int

    def get_config(self, topic_str):
        metadata_dict = self.get_metadata(topic_str)
        config_dict = metadata_dict["config"]
        #
        return config_dict

    def set_metadata(self, topic_str, metadata_dict):
        topic_dir_str = self.get_topic_abs_path_str(topic_str)
        self.write_dict_to_file(os.path.join(topic_dir_str, "metadata"), metadata_dict)

    # Groups

    def get_group_state(self, group_str):
        group_dict = self.get_group_dict(group_str)
        #
        state_str = group_dict["state"]
        #
        return state_str

    def get_group_dict(self, group_str):
        root_dir_str = self.storage_obj.root_dir()
        abs_path_file_str = os.path.join(root_dir_str, "groups", f"{group_str}")
        #
        group_dict = self.read_dict_from_file(abs_path_file_str)
        #
        return group_dict

    def set_group_dict(self, group_str, new_group_dict):
        root_dir_str = self.storage_obj.root_dir()
        abs_path_file_str = os.path.join(root_dir_str, "groups", f"{group_str}")
        #
        group_dict = self.read_dict_from_file(abs_path_file_str)
        #
        if "offsets" in new_group_dict:
            for topic_str, offsets_dict in new_group_dict["offsets"].items():
                if "offsets" not in group_dict:
                    group_dict["offsets"] = {}
                if topic_str not in group_dict["offsets"]:
                    group_dict["offsets"][topic_str] = {}
                #
                for partition_int, offset_int in offsets_dict.items():
                    group_dict["offsets"][topic_str][partition_int] = offset_int
        #
        if "last_update" in new_group_dict:
            group_dict["last_update"] = new_group_dict["last_update"]
        else:
            group_dict["last_update"] = get_millis()
        #
        if "state" in new_group_dict:
            group_dict["state"] = new_group_dict["state"]
        #
        self.write_dict_to_file(abs_path_file_str, group_dict)
        #
        return group_dict



from confluent_kafka import Producer

from kafi.kafka.kafka_producer import KafkaProducer

# Constants

CURRENT_TIME = 0
RD_KAFKA_PARTITION_UA = -1

#

class ClusterProducer(KafkaProducer):
    def __init__(self, cluster_obj, topic, **kwargs):
        super().__init__(cluster_obj, topic, **kwargs)
        #
        self.on_delivery_function = kwargs["on_delivery"] if "on_delivery" in kwargs else None
        #
        # Producer config
        #
        producer_config_dict = cluster_obj.kafka_config_dict.copy()
        #
        if "config" in kwargs:
            for key_str, value in kwargs["config"].items():
                producer_config_dict[key_str] = value
        #
        self.producer = Producer(producer_config_dict)

    def __del__(self):
        self.flush()

    #

    def close(self):
        self.flush()
        return self.topic_str

    #

    def flush(self):
        self.producer.flush(self.storage_obj.flush_timeout())
        #
        return self.topic_str

    def produce(self, value, **kwargs):
        key = kwargs["key"] if "key" in kwargs else None
        partition = kwargs["partition"] if "partition" in kwargs else RD_KAFKA_PARTITION_UA
        timestamp = kwargs["timestamp"] if "timestamp" in kwargs else CURRENT_TIME
        headers = kwargs["headers"] if "headers" in kwargs else None
        #
        flush_bool = kwargs["flush"] if "flush" in kwargs else False
        #
        value_list = value if isinstance(value, list) else [value]
        #
        key_list = key if isinstance(key, list) else [key for _ in value_list]
        #
        partition_int_list = partition if isinstance(partition, list) else [partition for _ in value_list]
        #
        timestamp_list = timestamp if isinstance(timestamp, list) else [timestamp for _ in value_list]
        #
        headers_list = headers if isinstance(headers, list) and all(self.storage_obj.is_headers(headers1) for headers1 in headers) and len(headers) == len(value_list) else [headers for _ in value_list]
        headers_str_bytes_tuple_list_list = [self.storage_obj.headers_to_headers_str_bytes_tuple_list(headers) for headers in headers_list]
        #
        for value, key, partition_int, timestamp, headers_str_bytes_tuple_list in zip(value_list, key_list, partition_int_list, timestamp_list, headers_str_bytes_tuple_list_list):
            key_str_or_bytes = self.serialize(key, True)
            value_str_or_bytes = self.serialize(value, False)
            #
            timestamp_int = timestamp[1] if isinstance(timestamp, tuple) else timestamp
            #
            self.producer.produce(self.topic_str, value_str_or_bytes, key_str_or_bytes, partition=partition_int, timestamp=timestamp_int, headers=headers_str_bytes_tuple_list, on_delivery=self.on_delivery_function)
            self.producer.poll(0) # https://stackoverflow.com/questions/62408128/buffererror-local-queue-full-in-python
            #
            self.written_counter_int += 1
        #
        if flush_bool:
            self.flush()
        #
        return self.written_counter_int



from kafi.storage import Storage

# Constants

ALL_MESSAGES = -1

#

class Kafka(Storage):
    def __init__(self, config_dir_str, config_name_str, mandatory_section_str_list, optional_section_str_list):
        super().__init__(config_dir_str, config_name_str, mandatory_section_str_list, optional_section_str_list)
        #
        self.config_dir_str = config_dir_str
        self.config_name_str = config_name_str
        #
        if "kafka" in mandatory_section_str_list:
            self.kafka_config_dict = self.config_dict["kafka"]
        else:
            self.kafka_config_dict = None
        #
        if "rest_proxy" in mandatory_section_str_list:
            self.rest_proxy_config_dict = self.config_dict["rest_proxy"]
        else:
            self.rest_proxy_config_dict = None
        #
        self.admin = None
        #
        # cluster config kafi section
        #
        if "flush.timeout" not in self.kafi_config_dict:
            self.flush_timeout(-1.0)
        else:
            self.flush_timeout(float(self.kafi_config_dict["flush.timeout"]))
        #
        if "retention.ms" not in self.kafi_config_dict:
            self.retention_ms(604800000)
        else:
            self.retention_ms(int(self.kafi_config_dict["retention.ms"]))
        #
        if "consume.timeout" not in self.kafi_config_dict:
            self.consume_timeout(5.0)
        else:
            self.consume_timeout(float(self.kafi_config_dict["consume.timeout"]))
        #
        if "session.timeout.ms" not in self.kafi_config_dict:
            self.session_timeout_ms(45000)
        else:
            self.session_timeout_ms(int(self.kafi_config_dict["session.timeout.ms"]))
        #
        if "block.num.retries" not in self.kafi_config_dict:
            self.block_num_retries(10)
        else:
            self.block_num_retries(int(self.kafi_config_dict["block.num.retries"]))
        #
        if "block.interval" not in self.kafi_config_dict:
            self.block_interval(0.5)
        else:
            self.block_interval(float(self.kafi_config_dict["block.interval"]))
        #
        # both cluster and restproxy kafi section
        #
        if "fetch.min.bytes" not in self.kafi_config_dict:
            self.fetch_min_bytes(-1)
        else:
            self.fetch_min_bytes(int(self.kafi_config_dict["fetch.min.bytes"]))
        #
        if "consumer.request.timeout.ms" not in self.kafi_config_dict:
            self.consumer_request_timeout_ms(1000)
        else:
            self.consumer_request_timeout_ms(int(self.kafi_config_dict["consumer.request.timeout.ms"]))
        #
        if "consume.num.attempts" not in self.kafi_config_dict:
            self.consume_num_attempts(3)
        else:
            self.consume_num_attempts(int(self.kafi_config_dict["consume.num.attempts"]))
        #
        if "requests.num.retries" not in self.kafi_config_dict:
            self.requests_num_retries(10)
        else:
            self.requests_num_retries(int(self.kafi_config_dict["requests.num.retries"]))

    # Cluster

    def flush_timeout(self, new_value=None): # float
        return self.get_set_config("flush.timeout", new_value)

    def retention_ms(self, new_value=None): # int
        return self.get_set_config("retention.ms", new_value)

    def consume_timeout(self, new_value=None): # float
        return self.get_set_config("consume.timeout", new_value)

    def enable_auto_commit(self, new_value=None): # bool
        return self.get_set_config("enable.auto.commit", new_value)

    def session_timeout_ms(self, new_value=None): # int
        return self.get_set_config("session.timeout.ms", new_value)

    def block_num_retries(self, new_value=None): # int
        return self.get_set_config("block.num.retries", new_value)

    def block_interval(self, new_value=None): # float
        return self.get_set_config("block.interval", new_value)

    # RestProxy

    def fetch_min_bytes(self, new_value=None): # int
        return self.get_set_config("fetch.min.bytes", new_value)

    def consumer_request_timeout_ms(self, new_value=None): # int
        return self.get_set_config("consumer.request.timeout.ms", new_value)

    def consume_num_attempts(self, new_value=None): # int
        return self.get_set_config("consume.num.attempts", new_value)

    def requests_num_retries(self, new_value=None): # int
        return self.get_set_config("requests.num.retries", new_value)

    # Brokers

    def brokers(self, pattern=None):
        return self.admin.brokers(pattern)
    
    def broker_config(self, pattern=None, config=None, **kwargs):
        return self.admin.broker_config(pattern, config, **kwargs)
    
    # ACLs

    def acls(self, restype="any", name=None, resource_pattern_type="any", principal=None, host=None, operation="any", permission_type="any"):
        return self.admin.acls(restype, name, resource_pattern_type, principal, host, operation, permission_type)

    def create_acl(self, restype="any", name=None, resource_pattern_type="any", principal=None, host=None, operation="any", permission_type="any"):
        return self.admin.create_acl(restype, name, resource_pattern_type, principal, host, operation, permission_type)
    
    def delete_acl(self, restype="any", name=None, resource_pattern_type="any", principal=None, host=None, operation="any", permission_type="any"):
        return self.admin.delete_acl(restype, name, resource_pattern_type, principal, host, operation, permission_type)



import pandas as pd

from kafi.functional import Functional
from kafi.helpers import explode_normalize

# Constants

ALL_MESSAGES = -1

#

class Pandas(Functional):
    def topic_to_df(self, topic, n=ALL_MESSAGES, **kwargs):
        #
        def foldl_function(acc, message_dict):
            # df = pd.DataFrame.from_records([message_dict["value"]])
            df = pd.json_normalize(message_dict["value"])
            if "explode" in kwargs and kwargs["explode"] == True:
                df = explode_normalize(df)
            #
            acc = pd.concat([acc, df], ignore_index=True)
            #
            return acc
        #

        (df,  _) = self.foldl(topic, foldl_function, pd.DataFrame(), n, **kwargs)
        #
        return df

    def df_to_topic(self, df, topic, n=ALL_MESSAGES, **kwargs):
        n_int = n
        #

        producer = self.producer(topic, **kwargs)
        counter_int = 0
        for _, row in df.iterrows():
            if n_int != ALL_MESSAGES:
                if counter_int >= n_int:
                    break
            #
            producer.produce(row.to_dict())
            #
            counter_int += 1
        producer.close()
        #
        return counter_int



import os
import sys

if os.path.basename(os.getcwd()) == "test":
    sys.path.insert(1, "..")
else:
    sys.path.insert(1, ".")

from test.test_single_storage_base import TestSingleStorageBase
from kafi.fs.azureblob.azureblob import AzureBlob

#

class Test(TestSingleStorageBase):
    def setUp(self):
        super().setUp()
        self.path_str = "test"

    def tearDown(self):
        super().tearDown()

    #

    def get_storage(self):
        a = AzureBlob("local")
        a.root_dir(self.path_str)
        #
        return a

    def is_ccloud(self):
        return False

    def test_acls(self):
        pass

    def test_brokers(self):
        pass

    def test_compact(self):
        pass



from kafi.storage import Storage
from kafi.files import Files

#

class FS(Storage):
    def __init__(self, config_dir_str, config_name_str, mandatory_section_str_list, optional_section_str_list):
        super().__init__(config_dir_str, config_name_str, mandatory_section_str_list, optional_section_str_list)
        #
        self.config_dir_str = config_dir_str
        self.config_name_str = config_name_str
        # local, azure_blob and s3
        if "local" in mandatory_section_str_list:
            self.local_config_dict = self.config_dict["local"]
            #
            if "root.dir" not in self.local_config_dict:
                self.root_dir(".")
            else:
                self.root_dir(str(self.local_config_dict["root.dir"]))
        else:
            self.local_config_dict = None
        # azure_blob
        if "azure_blob" in mandatory_section_str_list:
            self.azure_blob_config_dict = self.config_dict["azure_blob"]
            #
            if "root.dir" not in self.azure_blob_config_dict:
                self.root_dir("")
            else:
                self.root_dir(str(self.azure_blob_config_dict["root.dir"]))
            #
            if "container.name" not in self.azure_blob_config_dict:
                self.container_name("test")
            else:
                self.container_name(str(self.azure_blob_config_dict["container.name"]))
        else:
            self.azure_blob_config_dict = None
        # s3
        if "s3" in mandatory_section_str_list:
            self.s3_config_dict = self.config_dict["s3"]
            #
            if "root.dir" not in self.s3_config_dict:
                self.root_dir("")
            else:
                self.root_dir(str(self.s3_config_dict["root.dir"]))
            #
            if "bucket.name" not in self.s3_config_dict:
                self.bucket_name("test")
            else:
                self.bucket_name(str(self.s3_config_dict["bucket.name"]))
        else:
            self.s3_config_dict = None
        #
        self.admin = self.get_admin()

    # azure_blob

    def container_name(self, new_value=None): # str
        return self.get_set_config("container.name", new_value, dict=self.azure_blob_config_dict)

    # local
    
    def root_dir(self, new_value=None): # str
        return self.get_set_config("root.dir", new_value, dict=self.local_config_dict)

    # s3
    
    def bucket_name(self, new_value=None): # str
        return self.get_set_config("bucket.name", new_value, dict=self.s3_config_dict)
